# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/internal_parameter.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 14408 bytes
import contextlib

from Live import DeviceParameter
from past.builtins import unicode

from ..base import EventError, EventObject, Slot, clamp, listenable_property, liveobj_valid, nop


def identity(value, _parent):
    return value


def to_percentage_display(value):
    percentage = 100.0 * value
    percentage_str = "100"
    if percentage < 100.0:
        precision = 2 if percentage < 10.0 else 1
        format_str = "%." + str(precision) + "f"
        percentage_str = format_str % percentage
    return unicode(percentage_str + " %")


class InternalParameterBase(EventObject):
    is_enabled = True
    is_quantized = False

    def __init__(self, name=None, *a, **k):
        (super().__init__)(*a, **k)
        self._name = name
        self._state = DeviceParameter.ParameterState.enabled

    def _has_valid_parent(self):
        return liveobj_valid(self._parent)

    @property
    def canonical_parent(self):
        raise NotImplementedError

    @property
    def display_value(self):
        raise NotImplementedError

    @property
    def min(self):
        raise NotImplementedError

    @property
    def max(self):
        raise NotImplementedError

    @property
    def value(self):
        raise NotImplementedError

    @listenable_property
    def name(self):
        return self._name

    @property
    def original_name(self):
        return self._name

    @property
    def default_value(self):
        return self.min

    @listenable_property
    def automation_state(self):
        return DeviceParameter.AutomationState.none

    @listenable_property
    def state(self):
        return self._state

    @state.setter
    def state(self, new_state):
        self._state = new_state
        self.notify_state()

    @property
    def _live_ptr(self):
        return id(self)

    def __str__(self):
        return self.display_value


class InternalParameter(InternalParameterBase):
    __events__ = ("value",)

    def __init__(self, parent=None, display_value_conversion=None, *a, **k):
        (super().__init__)(*a, **k)
        self._value = 0.0
        self._parent = parent
        self.set_display_value_conversion(display_value_conversion)
        self.set_scaling_functions(None, None)

    def set_display_value_conversion(self, display_value_conversion):
        self._display_value_conversion = display_value_conversion or to_percentage_display
        self.notify_value()

    def set_scaling_functions(self, to_internal, from_internal):
        self._to_internal = to_internal or identity
        self._from_internal = from_internal or identity

    @property
    def canonical_parent(self):
        return self._parent

    def _get_value(self):
        if self._has_valid_parent():
            return self._from_internal(self.linear_value, self._parent)
        return self.min

    def _set_value(self, new_value):
        self.linear_value = self._to_internal(new_value, self._parent)

    value = property(_get_value, _set_value)

    def _get_linear_value(self):
        return self._value

    def _set_linear_value(self, new_value):
        if new_value != self._value:
            self._value = new_value
            self.notify_value()

    linear_value = property(_get_linear_value, _set_linear_value)

    @property
    def min(self):
        return 0.0

    @property
    def max(self):
        return 1.0

    @property
    def display_value(self):
        return self._display_value_conversion(self.value)


class PropertyHostMixin:
    def set_property_host(self, new_host):
        raise NotImplementedError


class WrappingParameter(InternalParameter, PropertyHostMixin):
    def __init__(
        self,
        property_host=None,
        source_property=None,
        from_property_value=None,
        to_property_value=None,
        display_value_conversion=nop,
        value_items=None,
        *a,
        **k,
    ):
        if value_items is None:
            value_items = []
        (super().__init__)(a, display_value_conversion=display_value_conversion, **k)
        self._property_host = property_host
        self._source_property = source_property
        self._value_items = value_items
        self.set_scaling_functions(to_property_value, from_property_value)
        self._property_slot = self.register_slot(
            Slot(listener=(self.notify_value), event_name=source_property, subject=(self._property_host)),
        )

    def set_property_host(self, new_host):
        self._property_host = new_host
        self._property_slot.subject = self._property_host

    def _get_property_value(self):
        if liveobj_valid(self._property_host):
            return getattr(self._property_host, self._source_property)
        return self.min

    def _get_value(self):
        try:
            if liveobj_valid(self._property_host):
                return self._from_internal(self._get_property_value(), self._property_host)
            return self.min
        except RuntimeError:
            return self.min

    def _set_value(self, new_value):
        if liveobj_valid(self._property_host):
            with contextlib.suppress(RuntimeError):
                setattr(self._property_host, self._source_property, self._to_internal(new_value, self._property_host))

    linear_value = property(_get_value, _set_value)
    value = property(_get_value, _set_value)

    @property
    def display_value(self):
        try:
            value = self._get_property_value()
            return unicode(self._display_value_conversion(value) if liveobj_valid(self._property_host) else "")
        except RuntimeError:
            return unicode()

    @property
    def is_quantized(self):
        return len(self._value_items) > 0

    @property
    def value_items(self):
        return self._value_items

    @property
    def short_value_items(self):
        return self.value_items


class EnumWrappingParameter(InternalParameterBase, PropertyHostMixin):
    is_enabled = True
    is_quantized = True

    def __init__(
        self,
        parent=None,
        index_property_host=None,
        values_host=None,
        values_property=None,
        index_property=None,
        value_type=int,
        to_index_conversion=None,
        from_index_conversion=None,
        *a,
        **k,
    ):
        (super().__init__)(*a, **k)
        self._parent = parent
        self._values_host = values_host
        self._index_property_host = index_property_host
        self._values_property = values_property
        self._index_property = index_property
        self._to_index = to_index_conversion or (lambda x: x)
        self._from_index = from_index_conversion or (lambda x: x)
        self.value_type = value_type
        self._index_property_slot = self.register_slot(index_property_host, self.notify_value, index_property)
        with contextlib.suppress(EventError):
            self.register_slot(self._values_host, self.notify_value_items, values_property)

    def set_property_host(self, new_host):
        self._index_property_host = new_host
        self._index_property_slot.subject = self._index_property_host

    @property
    def display_value(self):
        index = self._get_index()
        values = self._get_values()
        if index < len(values):
            return unicode(values[index])
        return unicode()

    @listenable_property
    def value_items(self):
        return self._get_values()

    @property
    def short_value_items(self):
        return self.value_items

    @listenable_property
    def value(self):
        return self._get_index()

    @value.setter
    def value(self, new_value):
        if new_value < 0 or new_value >= len(self._get_values()):
            raise IndexError
        self._set_index(new_value)

    def _get_values(self):
        if liveobj_valid(self._values_host):
            return getattr(self._values_host, self._values_property)
        return []

    def _get_index(self):
        return self._from_index(
            int(getattr(self._index_property_host, self._index_property))
            if liveobj_valid(self._index_property_host)
            else 0,
        )

    def _set_index(self, index):
        if liveobj_valid(self._index_property_host):
            index = self._to_index(index)
            setattr(self._index_property_host, self._index_property, self.value_type(index))

    @property
    def canonical_parent(self):
        return self._parent

    @property
    def max(self):
        return len(self.value_items) - 1

    @property
    def min(self):
        return 0


class RelativeInternalParameter(InternalParameter):
    __events__ = ("delta",)

    @property
    def default_value(self):
        return 0.5

    def _get_value(self):
        return self.default_value

    def _set_value(self, new_value):
        delta = new_value - self.value
        if delta != 0.0:
            self.notify_value()
            self.notify_delta(delta)

    value = property(_get_value, _set_value)
    linear_value = property(_get_value, _set_value)


class IntegerParameter(InternalParameter):
    def __init__(
        self,
        integer_value_host=None,
        integer_value_property_name=None,
        min_value=None,
        max_value=None,
        show_as_quantized=False,
        *a,
        **k,
    ):
        if "display_value_conversion" not in k:
            k["display_value_conversion"] = unicode
        (super().__init__)(*a, **k)
        self._integer_value_host = integer_value_host
        self._integer_value_property_name = integer_value_property_name
        self._min_value = min_value if min_value is not None else 0
        self._max_value = max_value if max_value is not None else 1
        self._show_as_quantized = show_as_quantized
        self._value = self._get_value()
        with contextlib.suppress(EventError):
            self.register_slot(integer_value_host, self.notify_value, integer_value_property_name)

    def _get_value(self):
        value = self._get_integer_value()
        if self._show_as_quantized:
            return self._index_from_value(value)
        return value

    def _set_value(self, new_value):
        if new_value != self._value:
            self._value = new_value
            self._set_integer_value(self._index_to_value(new_value) if self._show_as_quantized else new_value)

    value = property(_get_value, _set_value)

    def _get_linear_value(self):
        return self._value

    def _set_linear_value(self, new_value):
        if new_value != self._value:
            self._value = new_value
            if int(new_value) != self._get_integer_value():
                self._set_integer_value(new_value)
                self.notify_value()

    linear_value = property(_get_linear_value, _set_linear_value)

    def _get_integer_value(self):
        if liveobj_valid(self._integer_value_host):
            return getattr(self._integer_value_host, self._integer_value_property_name)
        return self._min_value

    def _set_integer_value(self, new_value):
        setattr(self._integer_value_host, self._integer_value_property_name, int(new_value))

    def _index_to_value(self, index):
        return clamp(self._min_value, index + self._min_value, self._max_value)

    def _index_from_value(self, value):
        return clamp(self._min_index, value - self._min_value, self._max_index)

    @property
    def _min_index(self):
        return 0

    @property
    def _max_index(self):
        return self._max_value - self._min_value

    @property
    def min(self):
        if self._show_as_quantized:
            return self._min_index
        return self._min_value

    @property
    def max(self):
        if self._show_as_quantized:
            return self._max_index
        return self._max_value

    @property
    def is_quantized(self):
        return self._show_as_quantized

    @property
    def value_items(self):
        return list(range(self._min_value, self._max_value + 1))

    @property
    def short_value_items(self):
        return self.value_items



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/elements_base.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 18352 bytes
from functools import partial

import Live

from ..base import chunks, flatten, is_iterable, recursive_map
from . import PrioritizedResource
from .elements import (
    ButtonElement,
    ButtonMatrixElement,
    ComboElement,
    DisplayLineElement,
    EncoderElement,
    SysexElement,
    SysexSendingButtonElement,
)


class MapMode:
    Absolute = Live.MidiMap.MapMode.absolute
    Absolute14Bit = Live.MidiMap.MapMode.absolute_14_bit
    AccelSignedBit = Live.MidiMap.MapMode.relative_signed_bit
    AccelSignedBit2 = Live.MidiMap.MapMode.relative_signed_bit2
    AccelBinaryOffset = Live.MidiMap.MapMode.relative_binary_offset
    AccelTwoCompliment = Live.MidiMap.MapMode.relative_two_compliment
    LinearSignedBit = Live.MidiMap.MapMode.relative_smooth_signed_bit
    LinearSignedBit2 = Live.MidiMap.MapMode.relative_smooth_signed_bit2
    LinearBinaryOffset = Live.MidiMap.MapMode.relative_smooth_binary_offset
    LinearTwoCompliment = Live.MidiMap.MapMode.relative_smooth_two_compliment


def create_button(identifier, name, **k):
    return ButtonElement(identifier, name=name, **k)


def create_sysex_sending_button(identifier, name, sysex_identifier, **k):
    return SysexSendingButtonElement(identifier, sysex_identifier, name=name, **k)


def create_encoder(identifier, name, **k):
    return EncoderElement(identifier, name=name, **k)


def create_sysex_element(identifier, name, send_message_generator=None, is_private=True, **k):
    return SysexElement(**k)


def create_combo_element(control=None, modifier=None, name=None, is_private=True):
    if not name:
        name = create_name_for_modified_control(control=control, modifier=modifier)
    return ComboElement(control=control, modifier=modifier, name=name, is_private=is_private)


def create_matrix_identifiers(start, stop, width=1, flip_rows=False):
    rows = list(chunks(list(range(start, stop)), width))
    if flip_rows:
        rows.reverse()
    return rows


def create_name_for_modified_control(control=None, modifier=None):
    modifier_name = modifier.name
    if modifier_name.lower().endswith("button"):
        modifier_name = modifier_name[:-7]
    preposition = "With" if any(c.isupper() for c in control.name) else "with"
    return f"{control.name}_{preposition}_{modifier_name}"


class ElementsBase:
    def __init__(self, global_channel=0, *a, **k):
        (super().__init__)(*a, **k)
        self._global_channel = global_channel

    def add_button(self, identifier, name, **k):
        attr_name = self._create_attribute_name(name)
        k["channel"] = k.get("channel", self._global_channel)
        setattr(self, attr_name, create_button(identifier, name, **k))

    def add_encoder(self, identifier, name, **k):
        attr_name = self._create_attribute_name(name)
        k["channel"] = k.get("channel", self._global_channel)
        setattr(self, attr_name, create_encoder(identifier, name, **k))

    def add_sysex_element(self, identifier, name, send_message_generator=None, **k):
        attr_name = self._create_attribute_name(name)
        setattr(self, attr_name, create_sysex_element(identifier, name, send_message_generator, **k))

    def add_sysex_display_line(self, identifier, name, send_message_generator=None, default_formatting=None, **k):
        command_element_name = name + "_Command"
        sysex_element = create_sysex_element(
            identifier,
            command_element_name,
            send_message_generator,
            optimized=True,
            **k,
        )
        setattr(self, self._create_attribute_name(command_element_name), sysex_element)
        display_line_kwargs = {"name": name, "display_fn": sysex_element.send_value}
        if default_formatting is not None:
            display_line_kwargs.update(default_formatting=default_formatting)
        setattr(self, self._create_attribute_name(name), DisplayLineElement(**display_line_kwargs))

    def add_element(self, name, element_factory, *a, **k):
        attr_name = self._create_attribute_name(name)
        setattr(self, attr_name, element_factory(a, name=name, **k))

    def add_modifier_button(self, identifier, name, *a, **k):
        (self.add_button)(identifier, name, *a, resource_type=PrioritizedResource, **k)

    def add_modified_control(
        self,
        control=None,
        modifier=None,
        name=None,
        element_factory=create_combo_element,
        is_private=True,
    ):
        if isinstance(control, ButtonMatrixElement):
            self._add_modified_matrix(control, modifier, name, element_factory, is_private)
        else:
            element = element_factory(control=control, modifier=modifier, name=name, is_private=is_private)
            attr_name = self._create_attribute_name(element.name)
            setattr(self, attr_name, element)

    def add_matrix(
        self,
        identifiers,
        base_name,
        channels=None,
        element_factory=None,
        name_factory=None,
        is_private=False,
        **k,
    ):
        channels = channels if channels is not None else self._global_channel
        if not is_iterable(channels):
            channels = [[channels] * len(row) for row in identifiers]

        def one_dimensional_name(name, x, _):
            return f"{name}_{x}"

        def two_dimensional_name(name, x, y):
            return f"{x}_{name}_{y}"

        name_factory = name_factory or (two_dimensional_name if len(identifiers) > 1 else one_dimensional_name)
        sub_element_name = base_name[:-1] if base_name.endswith("s") else base_name
        elements = [
            [
                element_factory(
                    identifier,
                    name=name_factory(sub_element_name, column, row),
                    channel=channel,
                    is_private=is_private,
                    **k,
                )
                for column, (identifier, channel) in enumerate(zip(inner_identifiers, inner_channels))
            ]
            for row, (inner_identifiers, inner_channels) in enumerate(zip(identifiers, channels))
        ]
        attr_name = self._create_attribute_name(base_name)
        self._add_raw_elements(attr_name, elements)
        setattr(self, attr_name, ButtonMatrixElement(name=base_name, rows=elements, is_private=is_private))

    def add_button_matrix(self, identifiers, base_name, channels=None, *a, **k):
        (self.add_matrix)(identifiers, base_name, *a, channels=channels, element_factory=create_button, **k)

    def add_encoder_matrix(self, identifiers, base_name, channels=None, *a, **k):
        (self.add_matrix)(identifiers, base_name, *a, channels=channels, element_factory=create_encoder, **k)

    def add_submatrix(self, matrix, name, columns=None, rows=None, is_private=True):
        if not columns:
            columns = (0, matrix.width() + 1)
        if not rows:
            rows = (0, matrix.height() + 1)
        submatrix = matrix.submatrix[columns[0] : columns[1], rows[0] : rows[1]]
        submatrix.name = name
        submatrix.is_private = is_private
        attr_name = self._create_attribute_name(name)
        setattr(self, attr_name, submatrix)

    def _add_modified_matrix(self, matrix, modifier, name, element_factory, is_private):
        modified_elements = recursive_map(
            partial(element_factory, modifier=modifier, is_private=is_private),
            matrix._orig_buttons,
        )
        if not name:
            name = create_name_for_modified_control(control=matrix, modifier=modifier)
        attr_name = self._create_attribute_name(name)
        self._add_raw_elements(attr_name, modified_elements)
        setattr(self, attr_name, ButtonMatrixElement(name=name, rows=modified_elements, is_private=is_private))

    def _add_raw_elements(self, base_name, elements):
        setattr(self, f"{base_name}_raw", list(flatten(elements)))

    def _create_attribute_name(self, name):
        return name.lower().replace(" ", "_")



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_MxDCore/MxDControlSurfaceAPI.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 7323 bytes
import contextlib
from collections import namedtuple

from ableton.v2.base import old_hasattr
from future.utils import string_types

from _MxDCore.ControlSurfaceWrapper import WrapperRegistry
from _MxDCore.LomTypes import get_control_surfaces, is_control_surface

RECEIVE_MIDI_TIMEOUT = 0.2


def midi_byte_to_int(byte):
    if isinstance(byte, string_types):
        return int(byte, 0)
    return byte


def midi_bytes_are_sysex(bytes):
    return len(bytes) > 3 and bytes[0] == 240 and bytes[-1] == 247


def check_has_mxd_midi_scheduler(lom_object, shown_name):
    if not old_hasattr(lom_object, "mxd_midi_scheduler"):
        raise AttributeError(f"object '{type(lom_object).__name__}' has no attribute '{shown_name}'")


class MxDControlSurfaceAPI:
    def __init__(self, mxdcore, *a, **k):
        (super().__init__)(*a, **k)
        self._mxdcore = mxdcore
        self._wrapper_registry = WrapperRegistry()

    @property
    def wrapper_registry(self):
        return self._wrapper_registry

    class MxDMidiOwner(namedtuple("Owner", "device_id object_id mxdcore")):
        def send_reply(self, selector, message):
            message_str = self.mxdcore.str_representation_for_object(message)
            selector_str = {
                "send_receive": "send_receive_sysex",
                "grab": "grab_midi",
                "release": "release_midi",
                "received": "received_midi",
            }.get(selector, str(selector))
            with contextlib.suppress(Exception):
                self.mxdcore.manager.send_message(
                    self.device_id,
                    self.object_id,
                    "obj_output",
                    '"' + selector_str + '"  ' + message_str,
                )

        def report_error(self, message):
            self.mxdcore._print_error(self.device_id, self.object_id, message)

        def is_expected_reply(self, message):
            return midi_bytes_are_sysex(message)

    def object_send_midi(self, device_id, object_id, lom_object, parameters):
        check_has_mxd_midi_scheduler(lom_object, "send_midi")
        midi_message = tuple(map(midi_byte_to_int, parameters[1:]))
        lom_object.mxd_midi_scheduler.send(self.MxDMidiOwner(device_id, object_id, self._mxdcore), midi_message)

    def object_send_receive_sysex(self, device_id, object_id, lom_object, parameters):
        check_has_mxd_midi_scheduler(lom_object, "send_receive_sysex")
        owner = self.MxDMidiOwner(device_id, object_id, self._mxdcore)
        has_timeout = len(parameters) > 2 and parameters[-2] == "timeout"
        timeout = parameters[-1] if has_timeout else RECEIVE_MIDI_TIMEOUT
        midi_parameters = parameters[1:-2] if has_timeout else parameters[1:]
        midi_message = tuple(map(midi_byte_to_int, midi_parameters))
        if midi_bytes_are_sysex(midi_message):
            lom_object.mxd_midi_scheduler.send_receive(owner, midi_message, timeout)
        else:
            self._mxdcore._print_error(device_id, object_id, "non-sysex passed to send_receive_sysex")

    def object_grab_midi(self, device_id, object_id, lom_object, parameters):
        check_has_mxd_midi_scheduler(lom_object, "grab_midi")
        lom_object.mxd_midi_scheduler.grab(self.MxDMidiOwner(device_id, object_id, self._mxdcore))

    def object_release_midi(self, device_id, object_id, lom_object, parameters):
        check_has_mxd_midi_scheduler(lom_object, "release_midi")
        lom_object.mxd_midi_scheduler.release(self.MxDMidiOwner(device_id, object_id, self._mxdcore))

    def release_control_surface_midi(self, device_id, object_id):
        for control_surface in map(self._wrapper_registry.wrap, get_control_surfaces()):
            if old_hasattr(control_surface, "mxd_midi_scheduler"):
                control_surface.mxd_midi_scheduler.disconnect(self.MxDMidiOwner(device_id, object_id, self._mxdcore))

    def object_get_control_names(self, device_id, object_id, lom_object, parameters):
        if not is_control_surface(lom_object):
            raise AttributeError(f"object '{type(lom_object).__name__}' has no attribute get_control_names")
        control_names = lom_object.control_names
        result = (
            "control_names %d\n" % len(control_names)
            + "".join([f"control {name}\n" for name in control_names])
            + "done"
        )
        self._mxdcore.manager.send_message(device_id, object_id, "obj_call_result", result)

    def object_get_control(self, device_id, object_id, lom_object, parameters):
        if not is_control_surface(lom_object):
            raise AttributeError(f"object '{type(lom_object).__name__}' has no attribute get_control")
        cs = lom_object
        name = parameters[1]
        control = cs.get_control_by_name(name)
        result_str = self._mxdcore.str_representation_for_object(control)
        self._mxdcore.manager.send_message(device_id, object_id, "obj_call_result", result_str)

    def _get_control_or_raise(self, cs, control_or_name, command):
        if not is_control_surface(cs):
            raise AttributeError(f"object '{type(cs).__name__}' has no attribute {command}")
        if not control_or_name:
            raise AttributeError(f"control id or name required for {command}")
        if isinstance(control_or_name, string_types):
            control = cs.get_control_by_name(control_or_name)
            assert control, f"{control_or_name} is not a control of '{type(cs).__name__}'"
        else:
            control = control_or_name
            if not cs.has_control(control):
                id_str = self._mxdcore.str_representation_for_object(control)
                raise AttributeError(f"'{type(control).__name__}' ({id_str}) is not a control of '{type(cs).__name__}'")
        return control

    def object_grab_control(self, device_id, object_id, lom_object, parameters):
        cs = lom_object
        control_or_name = parameters[1]
        control = self._get_control_or_raise(cs, control_or_name, "grab_control")
        cs.grab_control(control)

    def object_release_control(self, device_id, object_id, lom_object, parameters):
        cs = lom_object
        control_or_name = parameters[1]
        control = self._get_control_or_raise(cs, control_or_name, "release_control")
        cs.release_control(control)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/PhysicalDisplayElement.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 13797 bytes
from functools import partial
from itertools import chain, starmap

from future.utils import raise_
from past.utils import old_div

from . import Task
from .ControlElement import ControlElement
from .DisplayDataSource import adjust_string
from .LogicalDisplaySegment import LogicalDisplaySegment
from .NotifyingControlElement import NotifyingControlElement
from .Resource import ClientWrapper, ProxyResource, StackingResource
from .Util import const, first, group, in_range, lazy_attribute, maybe, nop, second, slice_size, slicer, to_slice


class _DisplayCentralResource(StackingResource):
    def __init__(self, root_display=None, *a, **k):
        (super().__init__)(*a, **k)
        self._root_display = root_display

    def _actual_owners(self):
        remaining_indexes = set(self._root_display.display_indexes)

        def filter_client(client):
            (display, _), __ = client
            result = remaining_indexes & display.display_indexes
            remaining_indexes.difference_update(display.display_indexes)
            return bool(result)

        return list(reversed(list(map(first, filter(filter_client, reversed(self._clients))))))


class DisplayError(Exception):
    pass


class DisplaySegmentationError(DisplayError):
    pass


class DisplayElement(ControlElement):
    class ProxiedInterface(ControlElement.ProxiedInterface):
        set_num_segments = nop
        set_data_sources = nop
        segment = const(LogicalDisplaySegment(1, nop))

    def __init__(self, width_in_chars=None, num_segments=1, *a, **k):
        (super().__init__)(*a, **k)
        self._width = width_in_chars
        self._logical_segments = []
        self.set_num_segments(num_segments)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.display_string!r}>"

    @property
    def display_string(self):
        return "".join(map(str, self._logical_segments))

    @property
    def width(self):
        return self._width

    @lazy_attribute
    def display_slice(self):
        return slicer(1)(nop)()[:]

    @lazy_attribute
    def display_indexes(self):
        return set(range(self._width))

    def disconnect(self):
        self._disconnect_segments()
        super().disconnect()

    def _disconnect_segments(self):
        for segment in self._logical_segments:
            segment.disconnect()

    @property
    def num_segments(self):
        return len(self._logical_segments)

    def set_num_segments(self, num_segments):
        width = self._width
        if not in_range(num_segments, 1, width) or width % num_segments != 0:
            raise_(
                DisplaySegmentationError,
                "Can not split display of size %d into %d segments" % (self.width, num_segments),
            )
        if num_segments != len(self._logical_segments):
            self._disconnect_segments()
            self._width_per_segment = old_div(width, num_segments)
            self._logical_segments = [
                LogicalDisplaySegment(self._width_per_segment, self.update) for _ in range(num_segments)
            ]

    def set_data_sources(self, sources):
        if not sources:
            self.set_num_segments(1)
            self.reset()
        else:
            self.set_num_segments(len(sources))
            for segment, source in zip(self._logical_segments, sources):
                segment.set_data_source(source)

    @property
    def segments(self):
        return tuple(self._logical_segments)

    def segment(self, index):
        return self._logical_segments[index]

    def reset(self):
        for segment in self._logical_segments:
            segment.set_data_source(None)

    def update(self):
        pass


class PhysicalDisplayElement(DisplayElement, NotifyingControlElement):
    _ascii_translations = {
        "0": 48,
        "1": 49,
        "2": 50,
        "3": 51,
        "4": 52,
        "5": 53,
        "6": 54,
        "7": 55,
        "8": 56,
        "9": 57,
        "A": 65,
        "B": 66,
        "C": 67,
        "D": 68,
        "E": 69,
        "F": 70,
        "G": 71,
        "H": 72,
        "I": 73,
        "J": 74,
        "K": 75,
        "L": 76,
        "M": 77,
        "N": 78,
        "O": 79,
        "P": 80,
        "Q": 81,
        "R": 82,
        "S": 83,
        "T": 84,
        "U": 85,
        "V": 86,
        "W": 87,
        "X": 88,
        "Y": 89,
        "Z": 90,
        "a": 97,
        "b": 98,
        "c": 99,
        "d": 100,
        "e": 101,
        "f": 102,
        "g": 103,
        "h": 104,
        "i": 105,
        "j": 106,
        "k": 107,
        "l": 108,
        "m": 109,
        "n": 110,
        "o": 111,
        "p": 112,
        "q": 113,
        "r": 114,
        "s": 115,
        "t": 116,
        "u": 117,
        "v": 118,
        "w": 119,
        "x": 120,
        "y": 121,
        "z": 122,
        "@": 64,
        " ": 32,
        "!": 33,
        '"': 34,
        "#": 35,
        "♯": 35,
        ".": 46,
        ",": 44,
        ":": 58,
        ";": 59,
        "?": 63,
        "<": 60,
        ">": 62,
        "[": 91,
        "]": 93,
        "_": 95,
        "-": 45,
        "|": 124,
        "&": 38,
        "^": 94,
        "~": 126,
        "`": 96,
        "'": 39,
        "%": 37,
        "(": 40,
        ")": 41,
        "/": 47,
        "\\": 92,
        "*": 42,
        "+": 43,
    }

    def __init__(self, *a, **k):
        self._central_resource = _DisplayCentralResource(
            root_display=self,
            on_received_callback=(self._on_central_resource_received),
            on_lost_callback=(self._on_central_resource_lost),
        )
        (super().__init__)(a, resource_type=self.nested_display_resource_factory(self), **k)
        self._translation_table = self._ascii_translations
        self._message_header = None
        self._message_tail = None
        self._message_clear_all = None
        self._message_to_send = None
        self._last_sent_message = None
        self._block_messages = False
        self._send_message_task = self._tasks.add(Task.run(self._send_message))
        self._send_message_task.kill()

    def nested_display_resource_factory(self, display):
        wrapper = ClientWrapper(wrap=(lambda c: (display, c)), unwrap=(partial(maybe(second))))
        return const(ProxyResource(proxied_resource=(self._central_resource), client_wrapper=wrapper))

    def _on_central_resource_received(self, display_client):
        display, client = display_client
        client.set_control_element(display, True)
        self.update()

    def _on_central_resource_lost(self, display_client):
        display, client = display_client
        client.set_control_element(display, False)
        self.update()

    @property
    @slicer(1)
    def subdisplay(self, char_slice):
        char_slice = to_slice(char_slice)
        return SubDisplayElement(sub_display_slice=char_slice, parent_display=self)

    def set_message_parts(self, header, tail):
        self._message_header = header
        self._message_tail = tail

    @property
    def message_header(self):
        return self._message_header

    @property
    def message_tail(self):
        return self._message_tail

    def set_clear_all_message(self, message):
        self._message_clear_all = message

    def set_translation_table(self, translation_table):
        self._translation_table = translation_table

    def set_block_messages(self, block):
        if block != self._block_messages:
            self._block_messages = block
        self.clear_send_cache()

    def display_message(self, message):
        if not self._block_messages:
            message = adjust_string(message, self._width)
            self._message_to_send = self._message_header + tuple(self._translate_string(message)) + self._message_tail
            self._request_send_message()

    def update(self):
        if len(self._logical_segments) > 0:
            if not self._block_messages:
                self._message_to_send = None
                self._request_send_message()

    def clear_send_cache(self):
        self._last_sent_message = None
        self._request_send_message()

    def reset(self):
        super().reset()
        if not self._block_messages:
            if self._message_clear_all is not None:
                self._message_to_send = self._message_clear_all
            else:
                self._message_to_send = tuple(
                    chain(self._message_header, self._translate_string(" " * int(self._width)), self._message_tail),
                )
            self._request_send_message()

    def send_midi(self, midi_bytes):
        if midi_bytes != self._last_sent_message:
            ControlElement.send_midi(self, midi_bytes)
            self._last_sent_message = midi_bytes

    def _request_send_message(self):
        self._send_message_task.restart()

    def _send_message(self):
        if not self._block_messages:
            if self._message_to_send is None:
                self._message_to_send = self._build_message(list(map(first, self._central_resource.owners)))
            self.send_midi(self._message_to_send)

    def _translate_char(self, char_to_translate):
        result = 63
        if char_to_translate in list(self._translation_table.keys()):
            result = self._translation_table[char_to_translate]
        else:
            result = self._translation_table["?"]
        return result

    def _translate_string(self, string):
        return list(map(self._translate_char, string))

    def _build_display_message(self, display):
        message_string = display.display_string
        segments = display._logical_segments
        width_per_segment = display._width_per_segment

        def wrap_segment_message(message, segment):
            return chain(segment.position_identifier(), self._translate_string(message))

        return chain(*starmap(wrap_segment_message, zip(group(message_string, width_per_segment), segments)))

    def _build_inner_message(self, displays):
        message = list(self._build_display_message(self))
        for display in displays:
            message[display.display_slice] = self._build_display_message(display)

        return message

    def _build_message(self, displays):
        return tuple(chain(self._message_header, self._build_inner_message(displays), self._message_tail))


class SubDisplayElement(DisplayElement):
    def __init__(self, sub_display_slice=slice(1), parent_display=None, *a, **k):
        (super().__init__)(
            a,
            width_in_chars=slice_size(sub_display_slice, parent_display.width),
            resource_type=parent_display.nested_display_resource_factory(self),
            **k,
        )
        self._sub_display_slice = sub_display_slice
        self._parent_display = parent_display

    @lazy_attribute
    def display_slice(self):
        return self._sub_display_slice

    @lazy_attribute
    def display_indexes(self):
        return set(range(*self._sub_display_slice.indices(self._parent_display.width)))

    def _is_visible(self):
        return self in list(map(first, self.resource.proxied_object.owners))

    def update(self):
        if self._is_visible():
            self._parent_display.update()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/display/text.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2792 bytes
from collections import UserString
from enum import Enum, auto
from typing import Callable, Tuple

from ...base import as_ascii
from ..elements import adjust_string


class Text(UserString):
    class ContentWidth:
        pass

    class Justification(Enum):
        LEFT = auto()
        CENTER = auto()
        RIGHT = auto()
        NONE = auto()

    def __init__(self, string="", justification=None, max_width=None):
        super().__init__(string)
        self.justification = justification
        self.max_width = max_width

    def as_ascii(self, adjust_string_fn: Callable[([str, int], str)] = adjust_string) -> Tuple[(int, ...)]:
        return tuple(as_ascii(self.as_string(adjust_string_fn=adjust_string_fn)))

    def as_string(self, adjust_string_fn: Callable[([str, int], str)] = adjust_string) -> str:
        max_width = self.max_width = (
            len(self) if (self.max_width is None or isinstance(self.max_width, Text.ContentWidth)) else (self.max_width)
        )
        justify = self.justification or Text.Justification.LEFT
        formatted = adjust_string_fn(str(self), max_width).rstrip() if self else ""
        if justify == Text.Justification.LEFT:
            return formatted.ljust(max_width)
        if justify == Text.Justification.CENTER:
            return formatted.center(max_width)
        if justify == Text.Justification.RIGHT:
            return formatted.rjust(max_width)
        return formatted



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/background.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2380 bytes
from functools import partial

from future.utils import itervalues

from ...base import nop
from ..component import Component


class BackgroundComponent(Component):
    def __init__(self, add_nop_listeners=False, *a, **k):
        (super().__init__)(*a, **k)
        self._add_nop_listeners = bool(add_nop_listeners)
        self._control_slots = {}
        self._control_map = {}

    def __getattr__(self, name):
        if len(name) > 4:
            if name[:4] == "set_":
                return partial(self._clear_control, name[4:])
        raise AttributeError(name)

    def _clear_control(self, name, control):
        slot = self._control_slots.get(name, None)
        if slot:
            del self._control_slots[name]
            self.disconnect_disconnectable(slot)
        if control:
            self._reset_control(control)
            self._control_map[name] = control
            if self._add_nop_listeners:
                self._control_slots[name] = self.register_slot(control, nop, "value")
        else:
            if name in self._control_map:
                del self._control_map[name]

    def _reset_control(self, control):
        control.reset()

    def update(self):
        super().update()
        if self.is_enabled():
            for control in itervalues(self._control_map):
                self._reset_control(control)


class ModifierBackgroundComponent(BackgroundComponent):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)

    def _reset_control(self, control):
        if len(control.resource.owners) > 1:
            control.set_light(True)
        else:
            control.reset()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/skin.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2497 bytes
from itertools import chain

from future.utils import iteritems

from ableton.v2.base import EventObject, const, depends, liveobj_valid
from ableton.v2.control_surface.elements.color import is_dynamic_color_factory


class SkinColorMissingError(Exception):
    pass


class DynamicColorNotAvailableError(Exception):
    msg = "In order to use dynamic colors, you need to inject the song while creating         the skin"


class Skin(EventObject):
    @depends(song=(const(None)))
    def __init__(self, colors=None, song=None, *a, **k):
        (super().__init__)(*a, **k)
        self._colors = {}
        self._factory_to_instance_map = {}
        if colors is not None:
            self._fill_colors(colors, song=song)

    def _fill_colors(self, colors, pathname="", song=None):
        if getattr(colors, "__bases__", None):
            for base in colors.__bases__:
                self._fill_colors(base, song=song, pathname=pathname)

        for k, v in iteritems(colors.__dict__):
            if k[:1] != "_":
                if callable(v):
                    self._fill_colors(v, (pathname + k + "."), song=song)
                else:
                    if is_dynamic_color_factory(v):
                        v = self._get_dynamic_color(v, song)
                    self._colors[pathname + k] = v

    def __getitem__(self, key):
        try:
            return self._colors[key]
        except KeyError:
            raise SkinColorMissingError(f"Skin color missing: {key!s}")

    def iteritems(self):
        return iteritems(self._colors)

    def _get_dynamic_color(self, color_factory, song):
        if not liveobj_valid(song):
            raise DynamicColorNotAvailableError
        else:
            if color_factory not in self._factory_to_instance_map:
                self._factory_to_instance_map[color_factory] = self._create_dynamic_color(color_factory, song)
            return self._factory_to_instance_map[color_factory]

    def _create_dynamic_color(self, color_factory, song):
        return self.register_disconnectable(color_factory.instantiate(song))


def merge_skins(*skins):
    skin = Skin()
    skin._colors = dict(chain(*(s._colors.items() for s in skins)))
    return skin



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/percussion_instrument_finder.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4038 bytes
from itertools import chain

import Live

from ..base import EventObject, listens_group, liveobj_changed, old_hasattr
from .device_chain_utils import find_instrument_devices, find_instrument_meeting_requirement
from .mode import Mode


class PercussionInstrumentFinder(Mode, EventObject):
    __events__ = ("instrument",)
    _drum_group = None
    _simpler = None

    def __init__(self, device_parent=None, is_enabled=True, *a, **k):
        (super().__init__)(*a, **k)
        self._is_enabled = is_enabled
        self._device_parent = None
        self.device_parent = device_parent

    @property
    def is_enabled(self):
        return self._is_enabled

    @is_enabled.setter
    def is_enabled(self, enabled):
        self._is_enabled = enabled
        self.update()

    def enter_mode(self):
        self.is_enabled = True

    def leave_mode(self):
        self.is_enabled = False

    @property
    def drum_group(self):
        return self._drum_group

    @property
    def sliced_simpler(self):
        return self._simpler

    @property
    def device_parent(self):
        return self._device_parent

    @device_parent.setter
    def device_parent(self, device_parent):
        if liveobj_changed(self._device_parent, device_parent):
            self._device_parent = device_parent
            self.update()

    @listens_group("devices")
    def __on_devices_changed(self, chain):
        self.device_parent.set_data("alternative_mode_locked", False)
        self.update()

    @listens_group("chains")
    def __on_chains_changed(self, chain):
        self.update()

    @listens_group("playback_mode")
    def __on_slicing_changed(self, _simpler):
        self.update()

    def update(self):
        if self.is_enabled:
            self._update_listeners()
            self._update_instruments()

    def _update_listeners(self):
        device_parent = self.device_parent
        devices = list(find_instrument_devices(device_parent))
        racks = list(filter(lambda d: d.can_have_chains, devices))
        simplers = list(filter(lambda d: old_hasattr(d, "playback_mode"), devices))
        chains = list(chain([device_parent], *[d.chains for d in racks]))
        self._PercussionInstrumentFinder__on_chains_changed.replace_subjects(racks)
        self._PercussionInstrumentFinder__on_devices_changed.replace_subjects(chains)
        self._PercussionInstrumentFinder__on_slicing_changed.replace_subjects(simplers)

    def _update_instruments(self):
        drum_group = find_drum_group_device(self.device_parent)
        simpler = find_sliced_simpler(self.device_parent)
        do_notify = liveobj_changed(drum_group, self._drum_group) or liveobj_changed(simpler, self._simpler)
        self._drum_group = drum_group
        self._simpler = simpler
        if do_notify:
            self.notify_instrument()


def find_drum_group_device(track_or_chain):
    def requirement(instrument):
        return instrument.can_have_drum_pads

    return find_instrument_meeting_requirement(requirement, track_or_chain)


def find_sliced_simpler(track_or_chain):
    def requirement(instrument):
        return getattr(instrument, "playback_mode", None) == Live.SimplerDevice.PlaybackMode.slicing

    return find_instrument_meeting_requirement(requirement, track_or_chain)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/ModesComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 28051 bytes
from functools import partial

from ableton.v2.base import old_hasattr

from . import Defaults, Task
from .CompoundComponent import CompoundComponent
from .ControlSurfaceComponent import ControlSurfaceComponent
from .Dependency import depends
from .Layer import LayerBase
from .Resource import StackingResource
from .SubjectSlot import subject_slot
from .Util import NamedTuple, infinite_context_manager, is_contextmanager, is_iterable, lazy_attribute


def tomode(thing):
    if thing is None:
        return Mode()
    if isinstance(thing, Mode):
        return thing
    if isinstance(thing, ControlSurfaceComponent):
        return ComponentMode(thing)
    if isinstance(thing, tuple):
        if len(thing) == 2:
            if isinstance(thing[0], ControlSurfaceComponent):
                if isinstance(thing[1], LayerBase):
                    return LayerMode(*thing)
            if callable(thing[0]):
                if callable(thing[1]):
                    mode = Mode()
                    mode.enter_mode, mode.leave_mode = thing
                    return mode
    if is_iterable(thing):
        return CompoundMode(*thing)
    if is_contextmanager(thing):
        return ContextManagerMode(thing)
    if callable(thing):
        mode = Mode()
        mode.enter_mode = thing
        return mode
    return thing


class Mode:
    def enter_mode(self):
        pass

    def leave_mode(self):
        pass

    def __enter__(self):
        self.enter_mode()

    def __exit__(self, *a):
        return self.leave_mode()


class ContextManagerMode(Mode):
    def __init__(self, context_manager=None, *a, **k):
        (super().__init__)(*a, **k)
        self._context_manager = context_manager

    def enter_mode(self):
        self._context_manager.__enter__()

    def leave_mode(self):
        self._context_manager.__exit__(None, None, None)

    def __exit__(self, exc_type, exc_value, traceback):
        return self._context_manager.__exit__(exc_type, exc_value, traceback)


def generator_mode(function):
    makecontext = infinite_context_manager(function)
    return lambda *a, **k: ContextManagerMode(makecontext(*a, **k))


class ComponentMode(Mode):
    def __init__(self, component=None, *a, **k):
        (super().__init__)(*a, **k)
        self._component = component

    def enter_mode(self):
        self._component.set_enabled(True)

    def leave_mode(self):
        self._component.set_enabled(False)


class LazyComponentMode(Mode):
    def __init__(self, component_creator=None, *a, **k):
        (super().__init__)(*a, **k)
        self._component_creator = component_creator

    @lazy_attribute
    def component(self):
        return self._component_creator()

    def enter_mode(self):
        self.component.set_enabled(True)

    def leave_mode(self):
        self.component.set_enabled(False)


class DisableMode(Mode):
    def __init__(self, component=None, *a, **k):
        (super().__init__)(*a, **k)
        self._component = component

    def enter_mode(self):
        self._component.set_enabled(False)

    def leave_mode(self):
        self._component.set_enabled(True)


class LayerModeBase(Mode):
    def __init__(self, component=None, layer=None, *a, **k):
        (super().__init__)(*a, **k)
        self._component = component
        self._layer = layer

    def _get_component(self):
        if callable(self._component):
            return self._component()
        return self._component


class LayerMode(LayerModeBase):
    def enter_mode(self):
        self._get_component().layer = self._layer

    def leave_mode(self):
        self._get_component().layer = None


class AddLayerMode(LayerModeBase):
    def enter_mode(self):
        self._layer.grab(self._get_component())

    def leave_mode(self):
        self._layer.release(self._get_component())


class CompoundMode(Mode):
    def __init__(self, *modes, **k):
        (super().__init__)(**k)
        self._modes = list(map(tomode, modes))

    def enter_mode(self):
        for mode in self._modes:
            mode.enter_mode()

    def leave_mode(self):
        for mode in reversed(self._modes):
            mode.leave_mode()


class MultiEntryMode(Mode):
    def __init__(self, mode=None, *a, **k):
        (super().__init__)(*a, **k)
        self._mode = tomode(mode)
        self._entry_count = 0

    def enter_mode(self):
        if self._entry_count == 0:
            self._mode.enter_mode()
        self._entry_count += 1

    def leave_mode(self):
        if self._entry_count == 1:
            self._mode.leave_mode()
        self._entry_count -= 1

    @property
    def is_entered(self):
        return self._entry_count > 0


class SetAttributeMode(Mode):
    def __init__(self, obj=None, attribute=None, value=None, *a, **k):
        (super().__init__)(*a, **k)
        self._obj = obj
        self._attribute = attribute
        self._old_value = None
        self._value = value

    def _get_object(self):
        if callable(self._obj):
            return self._obj()
        return self._obj

    def enter_mode(self):
        self._old_value = getattr(self._get_object(), self._attribute, None)
        setattr(self._get_object(), self._attribute, self._value)

    def leave_mode(self):
        if getattr(self._get_object(), self._attribute) == self._value:
            setattr(self._get_object(), self._attribute, self._old_value)


class DelayMode(Mode):
    @depends(parent_task_group=None)
    def __init__(self, mode=None, delay=None, parent_task_group=None, *a, **k):
        (super().__init__)(*a, **k)
        delay = delay if delay is not None else Defaults.MOMENTARY_DELAY
        self._mode = tomode(mode)
        self._mode_entered = False
        self._delay_task = parent_task_group.add(Task.sequence(Task.wait(delay), Task.run(self._enter_mode_delayed)))
        self._delay_task.kill()

    def _enter_mode_delayed(self):
        self._mode_entered = True
        self._mode.enter_mode()

    def enter_mode(self):
        self._delay_task.restart()

    def leave_mode(self):
        if self._mode_entered:
            self._mode.leave_mode()
            self._mode_entered = False
        self._delay_task.kill()


class ModeButtonBehaviour:
    def press_immediate(self, component, mode):
        pass

    def release_immediate(self, component, mode):
        pass

    def press_delayed(self, component, mode):
        pass

    def release_delayed(self, component, mode):
        pass

    def update_button(self, component, mode, selected_mode):
        button = component.get_mode_button(mode)
        groups = component.get_mode_groups(mode)
        selected_groups = component.get_mode_groups(selected_mode)
        button.set_light(mode == selected_mode or bool(groups & selected_groups))


class LatchingBehaviour(ModeButtonBehaviour):
    def press_immediate(self, component, mode):
        component.push_mode(mode)

    def release_immediate(self, component, mode):
        component.pop_unselected_modes()

    def release_delayed(self, component, mode):
        if len(component.active_modes) > 1:
            component.pop_mode(mode)


class ReenterBehaviour(LatchingBehaviour):
    def __init__(self, on_reenter=None, *a, **k):
        (super().__init__)(*a, **k)
        if on_reenter is not None:
            self.on_reenter = on_reenter

    def press_immediate(self, component, mode):
        was_active = component.selected_mode == mode
        super().press_immediate(component, mode)
        if was_active:
            self.on_reenter()

    def on_reenter(self):
        pass


class CancellableBehaviour(ModeButtonBehaviour):
    _previous_mode = None

    def press_immediate(self, component, mode):
        active_modes = component.active_modes
        groups = component.get_mode_groups(mode)
        can_cancel_mode = mode in active_modes or any(
            groups & component.get_mode_groups(other) for other in active_modes
        )
        if can_cancel_mode:
            if groups:
                component.pop_groups(groups)
            else:
                component.pop_mode(mode)
            self.restore_previous_mode(component)
        else:
            self.remember_previous_mode(component)
            component.push_mode(mode)

    def remember_previous_mode(self, component):
        self._previous_mode = component.active_modes[0] if component.active_modes else None

    def restore_previous_mode(self, component):
        if len(component.active_modes) == 0:
            if self._previous_mode is not None:
                component.push_mode(self._previous_mode)


class ImmediateBehaviour(ModeButtonBehaviour):
    def press_immediate(self, component, mode):
        component.selected_mode = mode


class AlternativeBehaviour(CancellableBehaviour):
    def __init__(self, alternative_mode=None, *a, **k):
        (super().__init__)(*a, **k)
        self._alternative_mode = alternative_mode

    def _check_mode_groups(self, component, mode):
        mode_groups = component.get_mode_groups(mode)
        alt_group = component.get_mode_groups(self._alternative_mode)
        return mode_groups and mode_groups & alt_group

    def release_delayed(self, component, mode):
        component.pop_groups(component.get_mode_groups(mode))
        self.restore_previous_mode(component)

    def press_delayed(self, component, mode):
        self.remember_previous_mode(component)
        component.push_mode(self._alternative_mode)

    def release_immediate(self, component, mode):
        super().press_immediate(component, mode)

    def press_immediate(self, component, mode):
        pass


class DynamicBehaviourMixin(ModeButtonBehaviour):
    def __init__(self, mode_chooser=None, *a, **k):
        (super().__init__)(*a, **k)
        self._mode_chooser = mode_chooser
        self._chosen_mode = None

    def press_immediate(self, component, mode):
        self._chosen_mode = self._mode_chooser() or mode
        super().press_immediate(component, self._chosen_mode)

    def release_delayed(self, component, mode):
        super().release_delayed(component, self._chosen_mode)

    def press_delayed(self, component, mode):
        super().press_delayed(component, self._chosen_mode)

    def release_immediate(self, component, mode):
        super().release_immediate(component, self._chosen_mode)


class ExcludingBehaviourMixin(ModeButtonBehaviour):
    def __init__(self, excluded_groups=None, *a, **k):
        if excluded_groups is None:
            excluded_groups = set()
        (super().__init__)(*a, **k)
        self._excluded_groups = set(excluded_groups)

    def is_excluded(self, component, selected):
        return bool(component.get_mode_groups(selected) & self._excluded_groups)

    def press_immediate(self, component, mode):
        if not self.is_excluded(component, component.selected_mode):
            super().press_immediate(component, mode)

    def release_delayed(self, component, mode):
        if not self.is_excluded(component, component.selected_mode):
            super().release_delayed(component, mode)

    def press_delayed(self, component, mode):
        if not self.is_excluded(component, component.selected_mode):
            super().press_delayed(component, mode)

    def release_immediate(self, component, mode):
        if not self.is_excluded(component, component.selected_mode):
            super().release_immediate(component, mode)

    def update_button(self, component, mode, selected_mode):
        if not self.is_excluded(component, selected_mode):
            super().update_button(component, mode, selected_mode)
        else:
            component.get_mode_button(mode).set_light("DefaultButton.Disabled")


class _ModeEntry(NamedTuple):
    mode = None
    groups = set()
    toggle_value = False
    subject_slot = None
    momentary_task = None


class ModesComponent(CompoundComponent):
    __subject_events__ = ("selected_mode",)
    momentary_toggle = False
    default_behaviour = LatchingBehaviour()

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._last_toggle_value = 0
        self._mode_toggle = None
        self._mode_toggle_task = self._tasks.add(Task.wait(Defaults.MOMENTARY_DELAY))
        self._mode_toggle_task.kill()
        self._mode_list = []
        self._mode_map = {}
        self._last_selected_mode = None
        self._mode_stack = StackingResource(self._do_enter_mode, self._do_leave_mode)
        self._shift_button = None

    def disconnect(self):
        self._mode_stack.release_all()
        super().disconnect()

    def set_shift_button(self, button):
        self._shift_button = button

    def _do_enter_mode(self, name):
        entry = self._mode_map[name]
        entry.mode.enter_mode()
        self._update_buttons(name)
        self.notify_selected_mode(name)

    def _do_leave_mode(self, name):
        self._mode_map[name].mode.leave_mode()
        if self._mode_stack.stack_size == 0:
            self._update_buttons(None)
            self.notify_selected_mode(None)

    def _get_selected_mode(self):
        return self._mode_stack.owner or self._last_selected_mode

    def _set_selected_mode(self, mode):
        if self.is_enabled():
            if mode is not None:
                self.push_mode(mode)
                self.pop_unselected_modes()
            else:
                self._mode_stack.release_all()
        else:
            self._last_selected_mode = mode

    selected_mode = property(_get_selected_mode, _set_selected_mode)

    @property
    def selected_groups(self):
        entry = self._mode_map.get(self.selected_mode, None)
        if entry:
            return entry.groups
        return set()

    @property
    def active_modes(self):
        return self._mode_stack.clients

    def push_mode(self, mode):
        self._mode_stack.grab(mode)

    def pop_mode(self, mode):
        self._mode_stack.release(mode)

    def pop_groups(self, groups):
        if not isinstance(groups, set):
            groups = set(groups)
        for client in self._mode_stack.clients:
            if self.get_mode_groups(client) & groups:
                self._mode_stack.release(client)

    def pop_unselected_modes(self):
        self._mode_stack.release_stacked()

    def on_enabled_changed(self):
        super().on_enabled_changed()
        if not self.is_enabled():
            self._last_selected_mode = self.selected_mode
            self._mode_stack.release_all()
        else:
            if self._last_selected_mode:
                self.push_mode(self._last_selected_mode)

    def update(self):
        super().update()
        self._update_buttons(self.selected_mode)

    def add_mode(self, name, mode_or_component, toggle_value=False, groups=None, behaviour=None):
        if groups is None:
            groups = set()
        if not isinstance(groups, set):
            groups = set(groups)
        mode = tomode(mode_or_component)
        task = self._tasks.add(
            Task.sequence(
                Task.wait(Defaults.MOMENTARY_DELAY),
                Task.run(
                    lambda: self._get_mode_behaviour(name).press_delayed(self, name),
                ),
            ),
        )
        task.kill()
        slot = self.register_slot(
            listener=(partial(self._on_mode_button_value, name)),
            event="value",
            extra_kws={"identify_sender": True},
        )
        self._mode_list.append(name)
        self._mode_map[name] = _ModeEntry(
            mode=mode,
            toggle_value=toggle_value,
            behaviour=behaviour,
            subject_slot=slot,
            momentary_task=task,
            groups=groups,
        )
        button_setter = "set_" + name + "_button"
        if not old_hasattr(self, button_setter):
            setattr(self, button_setter, partial(self.set_mode_button, name))

    def _get_mode_behaviour(self, name):
        entry = self._mode_map.get(name, None)
        return entry and (entry.behaviour) or (self.default_behaviour)

    def get_mode(self, name):
        entry = self._mode_map.get(name, None)
        return entry and entry.mode

    def get_mode_groups(self, name):
        entry = self._mode_map.get(name, None)
        if entry:
            return entry.groups
        return set()

    def set_toggle_button(self, button):
        if button:
            if self.is_enabled():
                button.reset()
        self._mode_toggle = button
        self._on_toggle_value.subject = button
        self._update_buttons(self.selected_mode)

    def set_mode_button(self, name, button):
        if button:
            if self.is_enabled():
                button.reset()
        self._mode_map[name].subject_slot.subject = button
        self._update_buttons(self.selected_mode)

    def get_mode_button(self, name):
        return self._mode_map[name].subject_slot.subject

    def _update_buttons(self, selected):
        if self.is_enabled():
            for name, entry in self._mode_map.items():
                if entry.subject_slot.subject is not None:
                    self._get_mode_behaviour(name).update_button(self, name, selected)

            if self._mode_toggle:
                entry = self._mode_map.get(selected)
                value = entry and entry.toggle_value
                self._mode_toggle.set_light(value)

    def _on_mode_button_value(self, name, value, sender):
        shift = self._shift_button and self._shift_button.is_pressed()
        if not shift:
            if self.is_enabled():
                behaviour = self._get_mode_behaviour(name)
                if sender.is_momentary():
                    entry = self._mode_map[name]
                    task = entry.momentary_task
                    if value:
                        behaviour.press_immediate(self, name)
                        task.restart()
                    else:
                        if task.is_killed:
                            behaviour.release_delayed(self, name)
                        else:
                            behaviour.release_immediate(self, name)
                            task.kill()
                else:
                    behaviour.press_immediate(self, name)
                    behaviour.release_immediate(self, name)

    @subject_slot("value")
    def _on_toggle_value(self, value):
        shift = self._shift_button and self._shift_button.is_pressed()
        if not shift:
            if self.is_enabled():
                if len(self._mode_list):
                    is_press = value and not self._last_toggle_value
                    is_release = not value and self._last_toggle_value
                    can_latch = self._mode_toggle_task.is_killed and self.selected_mode != self._mode_list[0]
                    if not self._mode_toggle.is_momentary() or is_press:
                        self.cycle_mode(1)
                        self._mode_toggle_task.restart()
                    else:
                        if is_release:
                            if self.momentary_toggle or can_latch:
                                self.cycle_mode(-1)
                    self._last_toggle_value = value

    def cycle_mode(self, delta=1):
        current_index = self._mode_list.index(self.selected_mode) if self.selected_mode else -delta
        current_index = (current_index + delta) % len(self._mode_list)
        self.selected_mode = self._mode_list[current_index]


class DisplayingModesComponent(ModesComponent):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._mode_data_sources = {}

    def add_mode(self, name, mode_or_component, data_source):
        super().add_mode(name, mode_or_component)
        self._mode_data_sources[name] = (data_source, data_source.display_string())

    def update(self):
        super().update()
        self._update_data_sources(self.selected_mode)

    def _do_enter_mode(self, name):
        super()._do_enter_mode(name)
        self._update_data_sources(name)

    def _update_data_sources(self, selected):
        if self.is_enabled():
            for name, (source, string) in self._mode_data_sources.items():
                source.set_display_string("*" + string if name == selected else string)


class EnablingModesComponent(ModesComponent):
    def __init__(self, component=None, toggle_value=False, disabled_value=False, *a, **k):
        (super().__init__)(*a, **k)
        component.set_enabled(False)
        self.add_mode("disabled", None, disabled_value)
        self.add_mode("enabled", component, toggle_value)
        self.selected_mode = "disabled"



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/auto_arm.py
# Compiled at: 2023-11-29 12:52:37
# Size of source mod 2**32: 4893 bytes
from ...base import listens, listens_group, task
from ...control_surface import Component


class AutoArmBase(Component):
    active_instances = []
    active_in_process_push_instances = []

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self.active_instances.append(self)
        self._AutoArmBase__on_control_surfaces_changed.subject = self.application
        self._update_implicit_arm_task = self._tasks.add(task.run(self._update_implicit_arm))

    def disconnect(self):
        self.active_instances.remove(self)
        if self.active_instances or self.active_in_process_push_instances:
            self._update_implicit_arm_task.restart()
        super().disconnect()

    def update(self):
        super().update()
        self._update_implicit_arm()

    @listens("control_surfaces")
    def __on_control_surfaces_changed(self):
        self._update_implicit_arm()

    def _update_implicit_arm(self):
        raise NotImplementedError


class AutoArmComponent(AutoArmBase):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._on_tracks_changed.subject = self.song
        self._on_exclusive_arm_changed.subject = self.song
        self._on_tracks_changed()
        self._on_selected_track_changed.subject = self.song.view

    @property
    def needs_restore_auto_arm(self):
        song = self.song
        exclusive_arm = song.exclusive_arm
        selected_track = song.view.selected_track
        return (
            self.can_auto_arm_track(selected_track)
            and not selected_track.arm
            and any(
                filter(
                    lambda track: (exclusive_arm or self.can_auto_arm_track(track))
                    and track.can_be_armed
                    and track.arm,
                    song.tracks,
                ),
            )
        )

    def track_can_be_armed(self, track):
        return track.can_be_armed and track.has_midi_input

    def can_auto_arm(self):
        return (
            any(instance.is_enabled() for instance in AutoArmComponent.active_instances)
            and not self.needs_restore_auto_arm
        )

    def can_auto_arm_track(self, track):
        return self.track_can_be_armed(track)

    def can_update_implicit_arm(self):
        return self.application.number_of_push_apps_running == 0 and (
            not AutoArmComponent.active_in_process_push_instances
            or self in AutoArmComponent.active_in_process_push_instances
        )

    @listens("selected_track")
    def _on_selected_track_changed(self):
        self.update()

    def _update_implicit_arm(self):
        self._update_implicit_arm_task.kill()
        if self.can_update_implicit_arm():
            song = self.song
            selected_track = song.view.selected_track
            can_auto_arm = self.can_auto_arm()
            for track in song.tracks:
                if self.track_can_be_armed(track):
                    track.implicit_arm = can_auto_arm and selected_track == track and self.can_auto_arm_track(track)

    @listens("tracks")
    def _on_tracks_changed(self):
        tracks = list(filter(lambda t: t.can_be_armed, self.song.tracks))
        self._on_arm_changed.replace_subjects(tracks)
        self._on_input_routing_type_changed.replace_subjects(tracks)
        self._on_frozen_state_changed.replace_subjects(tracks)

    @listens("exclusive_arm")
    def _on_exclusive_arm_changed(self):
        self.update()

    @listens_group("arm")
    def _on_arm_changed(self, track):
        self._update_implicit_arm_task.restart()

    @listens_group("input_routing_type")
    def _on_input_routing_type_changed(self, track):
        self._update_implicit_arm_task.restart()

    @listens_group("is_frozen")
    def _on_frozen_state_changed(self, track):
        self._update_implicit_arm_task.restart()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/mixer.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 13028 bytes
from functools import partial
from itertools import zip_longest

from ...base import EventObject, clamp, depends, forward_property, listenable_property, listens
from .. import Component
from ..controls import ButtonControl, MappedControl
from ..display import Renderable
from . import ChannelStripComponent

ASCII_A = 97


def send_letter_to_index(send_letter):
    return ord(send_letter) - ASCII_A


class SendIndexManager(EventObject, Renderable):
    @depends(song=None)
    def __init__(self, song=None, cycle_size=1, *a, **k):
        (super().__init__)(*a, **k)
        self._song = song
        self._cycle_size = cycle_size
        self._send_index = 0
        self._SendIndexManager__on_return_tracks_changed.subject = song
        self._SendIndexManager__on_return_tracks_changed()

    @listenable_property
    def num_sends(self):
        return len(self._song.return_tracks)

    @listenable_property
    def send_index(self):
        return self._send_index

    @send_index.setter
    def send_index(self, index):
        if index is not None:
            index = self._quantize_to_cycle_size(index)
        if self._send_index != index:
            self._send_index = index
            self.notify_send_index()

    def cycle_send_index(self, range_name="Send"):
        num_sends = self.num_sends
        if num_sends:
            if self._send_index < num_sends - self._cycle_size:
                self.send_index += self._cycle_size
            else:
                self.send_index = 0
            self._notify_send_range(range_name)

    def _quantize_to_cycle_size(self, value):
        value = clamp(value, 0, self.num_sends - 1)
        if value % self._cycle_size != 0:
            value = value // self._cycle_size * self._cycle_size
        return value

    def _update_send_index(self):
        num_sends = self.num_sends
        if self._send_index is not None:
            self.send_index = self._send_index if num_sends > 0 else None
        else:
            self.send_index = 0 if num_sends > 0 else None

    def _notify_send_range(self, range_name):
        first_letter = chr(ASCII_A + self._send_index).upper()
        last_letter = chr(ASCII_A + min(self._send_index + self._cycle_size, self.num_sends) - 1).upper()
        if self._cycle_size == 1 or first_letter == last_letter:
            self.notify(self.notifications.controlled_range, range_name, first_letter)
        else:
            self.notify(self.notifications.controlled_range, range_name, f"{first_letter} - {last_letter}")

    @listens("return_tracks")
    def __on_return_tracks_changed(self):
        self._update_send_index()
        self.notify_num_sends()


class MixerComponent(Component):
    prehear_volume_control = MappedControl()
    crossfader_control = MappedControl()
    cycle_send_index_button = ButtonControl(
        color="Mixer.CycleSendIndex",
        pressed_color="Mixer.CycleSendIndexPressed",
        disabled_color="Mixer.CycleSendIndexDisabled",
    )
    send_index = forward_property("_send_index_manager")("send_index")

    @depends(session_ring=None, target_track=None)
    def __init__(
        self,
        name="Mixer",
        session_ring=None,
        target_track=None,
        channel_strip_component_type=None,
        target_can_be_master=True,
        *a,
        **k,
    ):
        (super().__init__)(a, name=name, **k)
        self._target_track = target_track
        self._provider = session_ring
        self._MixerComponent__on_offset_changed.subject = self._provider
        self.register_slot(self._provider, self._reassign_tracks, "tracks")
        self._send_controls = None
        self._channel_strip_component_type = channel_strip_component_type or ChannelStripComponent
        self._channel_strips = [self._create_channel_strip() for _ in range(self._provider.num_tracks)]
        self._reassign_tracks()
        self._target_can_be_master = target_can_be_master
        self._target_strip = self._create_channel_strip(is_target=True)
        self._MixerComponent__on_target_track_changed.subject = self._target_track
        self._update_target_strip()
        self._master_strip = self._create_channel_strip(is_master=True)
        self._master_strip.set_track(self.song.master_track)
        self._send_index_manager = self.register_disconnectable(SendIndexManager())
        self.register_slot(self._send_index_manager, self._on_send_index_changed, "send_index")
        self.register_slot(self._send_index_manager, self._update_cycle_send_index_button, "num_sends")

    def channel_strip(self, index):
        return self._channel_strips[index]

    @property
    def master_strip(self):
        return self._master_strip

    @property
    def target_strip(self):
        return self._target_strip

    def set_prehear_volume_control(self, control):
        self.prehear_volume_control.set_control_element(control)

    def set_crossfader_control(self, control):
        self.crossfader_control.set_control_element(control)

    def set_cycle_send_index_button(self, button):
        self.cycle_send_index_button.set_control_element(button)

    def set_shift_button(self, button):
        for strip in self._channel_strips:
            strip.shift_button.set_control_element(button)

    def set_send_controls(self, controls):
        self._send_controls = controls
        for strip, control in zip_longest(self._channel_strips, controls or []):
            if self._send_index_manager.send_index is None:
                strip.send_controls.set_control_element((control,))
            else:
                strip.send_controls.set_control_element((None,) * self._send_index_manager.send_index + (control,))
                strip.update()

    def __getattr__(self, name):
        if name.startswith("set_master_track"):
            return partial(
                self._set_master_or_target_strip_control,
                self._master_strip,
                name.replace("set_master_track_", ""),
            )
        if name.startswith("set_target_track"):
            if "send" in name:
                if not name.endswith("send_controls"):
                    return partial(
                        self._set_target_strip_indexed_send_control,
                        send_letter_to_index(name.split("_")[-2]),
                    )
                return partial(
                    self._set_master_or_target_strip_control,
                    self._target_strip,
                    name.replace("set_target_track_", ""),
                )
        if "send" in name:
            if not name.endswith("send_controls"):
                return partial(self._set_indexed_send_controls, send_letter_to_index(name.split("_")[-2]))
            if name.startswith("set"):
                return partial(self._set_strip_controls, name[4:-1])
        raise AttributeError

    @staticmethod
    def _set_master_or_target_strip_control(strip, name, control):
        getattr(strip, name).set_control_element(control)

    def _set_target_strip_indexed_send_control(self, send_index, control):
        self._target_strip.set_indexed_send_control(control, send_index)

    def _set_indexed_send_controls(self, send_index, controls):
        for strip, control in zip_longest(self._channel_strips, controls or []):
            strip.set_indexed_send_control(control, send_index)

    def _set_strip_controls(self, name, controls):
        for strip, control in zip_longest(self._channel_strips, controls or []):
            getattr(strip, name).set_control_element(control)

    @cycle_send_index_button.pressed
    def cycle_send_index_button(self, _):
        self.cycle_send_index()

    def cycle_send_index(self):
        self._send_index_manager.cycle_send_index()

    def _on_send_index_changed(self):
        self.set_send_controls(self._send_controls)

    @listens("offset")
    def __on_offset_changed(self, *_):
        self._reassign_tracks()

    def _reassign_tracks(self):
        for track, channel_strip in zip(self._provider.tracks, self._channel_strips):
            channel_strip.set_track(track)

    @listens("target_track")
    def __on_target_track_changed(self):
        self._update_target_strip()

    def _update_target_strip(self):
        target_track = self._target_track.target_track
        if self._target_can_be_master or target_track != self.song.master_track:
            self._target_strip.set_track(target_track)
        else:
            self._target_strip.set_track(None)

    def update(self):
        super().update()
        if self.is_enabled():
            master_track = self.song.master_track
            self.prehear_volume_control.mapped_parameter = master_track.mixer_device.cue_volume
            self.crossfader_control.mapped_parameter = master_track.mixer_device.crossfader
        else:
            self.prehear_volume_control.mapped_parameter = None
            self.crossfader_control.mapped_parameter = None

    def _update_cycle_send_index_button(self):
        self.cycle_send_index_button.enabled = self._send_index_manager.num_sends

    def _create_channel_strip(self, is_master=False, is_target=False):
        return self._channel_strip_component_type(parent=self)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/clip_creator.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1178 bytes
import Live

_Q = Live.Song.Quantization


class ClipCreator:
    grid_quantization = None
    is_grid_triplet = False
    fixed_length = 8
    legato_launch = True

    def create(self, slot, length=None, launch_quantization=None, legato_launch=None):
        if length is None:
            length = self.fixed_length
        slot.create_clip(length)
        should_legato_launch = self.legato_launch if legato_launch is None else legato_launch
        if self.grid_quantization is not None:
            slot.clip.view.grid_quantization = self.grid_quantization
            slot.clip.view.grid_is_triplet = self.is_grid_triplet
        if launch_quantization is None or should_legato_launch:
            launch_quantization = _Q.q_no_q
        slot.fire(force_legato=should_legato_launch, launch_quantization=launch_quantization)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/control/__init__.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1797 bytes
from .button import ButtonControl, ButtonControlBase, DoubleClickContext, PlayableControl
from .control import (
    Connectable,
    Control,
    ControlManager,
    InputControl,
    SendValueControl,
    SendValueMixin,
    control_color,
    control_event,
    forward_control,
)
from .control_list import ControlList, MatrixControl, RadioButtonGroup, control_list, control_matrix
from .encoder import (
    EncoderControl,
    ListIndexEncoderControl,
    ListValueEncoderControl,
    SendValueEncoderControl,
    StepEncoderControl,
)
from .mapped import MappedControl, MappedSensitivitySettingControl, is_internal_parameter
from .radio_button import RadioButtonControl
from .sysex import ColorSysexControl
from .text_display import ConfigurableTextDisplayControl, TextDisplayControl
from .toggle_button import ToggleButtonControl

__all__ = (
    "ButtonControl",
    "ButtonControlBase",
    "ColorSysexControl",
    "ConfigurableTextDisplayControl",
    "Connectable",
    "Control",
    "ControlList",
    "ControlManager",
    "DoubleClickContext",
    "EncoderControl",
    "InputControl",
    "ListIndexEncoderControl",
    "ListValueEncoderControl",
    "MappedControl",
    "MappedSensitivitySettingControl",
    "MatrixControl",
    "PlayableControl",
    "RadioButtonControl",
    "RadioButtonGroup",
    "SendValueControl",
    "SendValueEncoderControl",
    "SendValueMixin",
    "StepEncoderControl",
    "TextDisplayControl",
    "ToggleButtonControl",
    "TouchableControl",
    "control_color",
    "control_event",
    "control_list",
    "control_matrix",
    "forward_control",
    "is_internal_parameter",
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/session_ring_selection_linking.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3319 bytes
from ..base import EventObject, depends, index_if, listens
from ..live import scene_index


class SessionRingSelectionLinking(EventObject):
    @depends(song=None, session_ring=None)
    def __init__(
        self,
        song=None,
        session_ring=None,
        selection_changed_notifier=None,
        link_to_track_selection=True,
        link_to_scene_selection=False,
        *a,
        **k,
    ):
        (super().__init__)(*a, **k)
        self.song = song
        self._session_ring = session_ring
        if link_to_track_selection:
            self._SessionRingSelectionLinking__on_track_selection_scrolled.subject = selection_changed_notifier
        if link_to_scene_selection:
            self._SessionRingSelectionLinking__on_scene_selection_scrolled.subject = selection_changed_notifier

    @listens("track_selection_scrolled")
    def __on_track_selection_scrolled(self):
        if self.song.view.selected_track not in self._session_ring.tracks_to_use():
            return
        track_index = index_if(lambda t: t == self.song.view.selected_track, self._session_ring.tracks_to_use())
        self._link_session_ring_with_minimal_travel("track", track_index)

    @listens("scene_selection_scrolled")
    def __on_scene_selection_scrolled(self):
        self._link_session_ring_with_minimal_travel("scene", scene_index())

    def _link_session_ring_with_minimal_travel(self, axis_name, current_index):
        ring_axis_offset = getattr(self._session_ring, f"{axis_name}_offset")
        ring_axis_size = getattr(self._session_ring, f"num_{axis_name}s")
        ending_ring_index = ring_axis_offset + ring_axis_size - 1
        offset_start = current_index - ring_axis_offset
        offset_end = current_index - ending_ring_index
        adjustment = min(0, offset_start) + max(0, offset_end)
        new_offset = ring_axis_offset + adjustment
        if new_offset != ring_axis_offset:
            track_offset = new_offset
            scene_offset = self._session_ring.scene_offset
            if axis_name == "scene":
                track_offset = self._session_ring.track_offset
                scene_offset = new_offset
            self._session_ring.set_offsets(track_offset, scene_offset)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/DrumRackComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2093 bytes
from .ControlSurfaceComponent import ControlSurfaceComponent
from .Dependency import depends
from .Util import product

NUM_PADS_X = 4
NUM_PADS_Y = 4


def _validate_matrix(matrix):
    if matrix.width() > NUM_PADS_X or matrix.height() > NUM_PADS_Y:
        raise RuntimeError("The provided button matrix should not be bigger than %dx%d" % (NUM_PADS_X, NUM_PADS_Y))


class DrumRackComponent(ControlSurfaceComponent):
    @depends(set_pad_translations=None, request_rebuild_midi_map=None)
    def __init__(self, set_pad_translations=None, request_rebuild_midi_map=None, *a, **k):
        (super().__init__)(*a, **k)
        self._set_pad_translations = set_pad_translations
        self._request_rebuild_midi_map = request_rebuild_midi_map

    def _create_and_set_pad_translations(self, matrix):
        def create_translation_entry(y_x):
            y, x = y_x
            button = matrix.get_button(x, y)
            return (
                x,
                y + NUM_PADS_Y - matrix.height(),
                button.message_identifier() if button is not None else 0,
                button.message_channel() if button is not None else 0,
            )

        translations = list(map(create_translation_entry, product(range(matrix.height()), range(matrix.width()))))
        self._set_pad_translations(tuple(translations))

    def set_pads(self, matrix):
        if matrix is not None:
            _validate_matrix(matrix)
            self._create_and_set_pad_translations(matrix)
            for button in filter(None, matrix):
                button.suppress_script_forwarding = True

        else:
            self._set_pad_translations(None)
        self._request_rebuild_midi_map()



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class AudioToMidiType:
    """
    Class that represent an enumeration of values for AudioToMidiType
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    harmony_to_midi = 0
    melody_to_midi = 1
    drums_to_midi = 2


def audio_to_midi_clip(audio_clip: object, audio_to_midi_type: int) -> None:
    """
    audio_to_midi_clip( (Song)song, (Clip)audio_clip, (int)audio_to_midi_type) -> None :
        Creates a MIDI clip in a new MIDI track with the notes extracted from the given
        audio_clip. The `audio_to_midi_type` decides which algorithm is used in
        the process. Raises error when called with an inconvertible clip or invalid
        `audio_to_midi_type`.

        C++ signature :
            void audio_to_midi_clip(TPyHandle<ASong>,TPyHandle<AClip>,int)
    """


def create_drum_rack_from_audio_clip(audio_clip: object) -> None:
    """
    create_drum_rack_from_audio_clip( (Song)song, (Clip)audio_clip) -> None :
        Creates a new track with a drum rack with a simpler on the first pad with
        the specified audio clip.

        C++ signature :
            void create_drum_rack_from_audio_clip(TPyHandle<ASong>,TPyHandle<AClip>)
    """


def create_midi_track_from_drum_pad(drum_pad: object) -> None:
    """
    create_midi_track_from_drum_pad( (Song)song, (DrumPad)drum_pad) -> None :
        Creates a new Midi track containing the specified Drum Pad's device chain.

        C++ signature :
            void create_midi_track_from_drum_pad(TPyHandle<ASong>,TPyHandle<ADrumGroupDevicePad>)
    """


def create_midi_track_with_simpler(audio_clip: object) -> None:
    """
    create_midi_track_with_simpler( (Song)song, (Clip)audio_clip) -> None :
        Creates a new Midi track with a simpler including the specified audio clip.

        C++ signature :
            void create_midi_track_with_simpler(TPyHandle<ASong>,TPyHandle<AClip>)
    """


def is_convertible_to_midi(audio_clip: object) -> bool:
    """
    is_convertible_to_midi( (Song)song, (Clip)audio_clip) -> bool :
        Returns whether `audio_clip` can be converted to MIDI.
        Raises error when called with a MIDI clip

        C++ signature :
            bool is_convertible_to_midi(TPyHandle<ASong>,TPyHandle<AClip>)
    """


def move_devices_on_track_to_new_drum_rack_pad(track_index: int) -> LomObject:
    """
    move_devices_on_track_to_new_drum_rack_pad( (Song)song, (int)track_index) -> LomObject :
        Moves the entire device chain of the track according to the track index
        onto the C1 (note 36) drum pad of a new drum rack in a new track.If the track associated with the track index does not contain any devices
        nothing changes (i.e. a new track and new drum rack are not created).

        C++ signature :
            TWeakPtr<TPyHandleBase> move_devices_on_track_to_new_drum_rack_pad(TPyHandle<ASong>,int)
    """


def sliced_simpler_to_drum_rack(simpler: object) -> None:
    """
    sliced_simpler_to_drum_rack( (Song)song, (SimplerDevice)simpler) -> None :
        Converts the Simpler into a Drum Rack, assigning each slice to a drum pad.
        Calling it on a non-sliced simpler raises an error.

        C++ signature :
            void sliced_simpler_to_drum_rack(TPyHandle<ASong>,TSimplerDevicePyHandle)
    """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/session_navigation.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 7277 bytes
from functools import partial

from ...base import clamp, depends, listens
from .. import Component
from . import Scrollable, ScrollComponent


class SessionRingScroller(Scrollable):
    def __init__(
        self,
        session_ring,
        respect_borders,
        snap_track_offset=False,
        scroll_scenes=False,
        page_size=1,
        *a,
        **k,
    ):
        (super().__init__)(*a, **k)
        self.session_ring = session_ring
        self.respect_borders = respect_borders
        self.snap_track_offset = snap_track_offset
        self.page_size = page_size
        can_scroll = self._can_scroll_scenes if scroll_scenes else self._can_scroll_tracks
        self.can_scroll_up = partial(can_scroll, -1)
        self.can_scroll_down = partial(can_scroll, 1)
        self._do_scroll = self._scroll_scenes if scroll_scenes else self._scroll_tracks

    def scroll_up(self):
        if self.can_scroll_up():
            self._do_scroll(-self.page_size)

    def scroll_down(self):
        if self.can_scroll_down():
            self._do_scroll(self.page_size)

    def _max_track_offset(self):
        if self.respect_borders:
            return len(self.session_ring.tracks_to_use()) - self.session_ring.num_tracks
        return len(self.session_ring.tracks_to_use()) - 1

    def _max_scene_offset(self):
        if self.respect_borders:
            return len(self.session_ring.scenes_to_use()) - self.session_ring.num_scenes
        return len(self.session_ring.scenes_to_use()) - 1

    def _can_scroll_tracks(self, delta):
        offset = self.session_ring.track_offset
        if self.snap_track_offset:
            if delta > 0:
                return offset < len(self.session_ring.tracks_to_use()) - self.page_size
        return delta < 0 < offset or offset + delta in range(self._max_track_offset() + 1)

    def _can_scroll_scenes(self, delta):
        offset = self.session_ring.scene_offset
        return delta < 0 < offset or offset + delta in range(self._max_scene_offset() + 1)

    def _scroll_tracks(self, delta):
        new_offset = self.session_ring.track_offset + delta
        self.session_ring.set_offsets(clamp(new_offset, 0, self._max_track_offset()), self.session_ring.scene_offset)

    def _scroll_scenes(self, delta):
        new_offset = self.session_ring.scene_offset + delta
        self.session_ring.set_offsets(self.session_ring.track_offset, clamp(new_offset, 0, self._max_scene_offset()))


class SessionNavigationComponent(Component):
    @depends(session_ring=None)
    def __init__(
        self,
        name="Session_Navigation",
        session_ring=None,
        respect_borders=False,
        snap_track_offset=False,
        *a,
        **k,
    ):
        (super().__init__)(a, name=name, **k)

        def scroller_factory(**k):
            return ScrollComponent(
                SessionRingScroller(session_ring, respect_borders, **k),
                parent=self,
                scroll_skin_name="Session.Navigation",
            )

        self._vertical_banking = scroller_factory(scroll_scenes=True)
        self._horizontal_banking = scroller_factory()
        self._vertical_paging = scroller_factory(scroll_scenes=True, page_size=(session_ring.num_scenes))
        self._horizontal_paging = scroller_factory(
            page_size=(session_ring.num_tracks),
            snap_track_offset=snap_track_offset,
        )
        self.register_slot(self.song, self._update_vertical, "scenes")
        self.register_slot(session_ring, self._update_horizontal, "tracks")
        self._SessionNavigationComponent__on_offset_changed.subject = session_ring

    def set_up_button(self, button):
        self._vertical_banking.set_scroll_up_button(button)

    def set_down_button(self, button):
        self._vertical_banking.set_scroll_down_button(button)

    def set_left_button(self, button):
        self._horizontal_banking.set_scroll_up_button(button)

    def set_right_button(self, button):
        self._horizontal_banking.set_scroll_down_button(button)

    def set_page_up_button(self, page_up_button):
        self._vertical_paging.set_scroll_up_button(page_up_button)

    def set_page_down_button(self, page_down_button):
        self._vertical_paging.set_scroll_down_button(page_down_button)

    def set_page_left_button(self, page_left_button):
        self._horizontal_paging.set_scroll_up_button(page_left_button)

    def set_page_right_button(self, page_right_button):
        self._horizontal_paging.set_scroll_down_button(page_right_button)

    def set_vertical_encoder(self, control):
        self._vertical_banking.set_scroll_encoder(control)

    def set_horizontal_encoder(self, control):
        self._horizontal_banking.set_scroll_encoder(control)

    def _update_vertical(self):
        if self.is_enabled():
            self._vertical_banking.update()
            self._vertical_paging.update()

    def _update_horizontal(self):
        if self.is_enabled():
            self._horizontal_banking.update()
            self._horizontal_paging.update()

    @listens("offset")
    def __on_offset_changed(self, *_):
        self._update_vertical()
        self._update_horizontal()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Generic/GenericScript.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 16711 bytes
from functools import partial

import Live
from _Framework import (
    ButtonElement,
    ButtonMatrixElement,
    ControlSurface,
    DeviceComponent,
    EncoderElement,
    Layer,
    TransportComponent,
)
from _Framework.InputControlElement import MIDI_CC_TYPE
from _Framework.Util import NamedTuple
from future.utils import iteritems

from .SpecialMixerComponent import SpecialMixerComponent


def is_valid_midi_channel(integer):
    return 0 <= integer < 16


def is_valid_midi_identifier(integer):
    return 0 <= integer < 128


def has_specification_for(control, specifications):
    return is_valid_midi_identifier(specifications.get(control, -1))


class TransportButton(NamedTuple):
    control_name = ""
    layer_name = ""


TRANSPORT_BUTTON_SPECIFICATIONS = {
    "Stop": TransportButton(control_name="Stop", layer_name="stop"),
    "Play": TransportButton(control_name="Play", layer_name="play"),
    "Rec": TransportButton(control_name="Record", layer_name="record"),
    "SessionRec": TransportButton(control_name="Session_Record", layer_name="session_record"),
    "Over": TransportButton(control_name="Arrangement_Overdub", layer_name="arrangement_overdub"),
    "Metro": TransportButton(control_name="Metronome", layer_name="metronome"),
    "Loop": TransportButton(control_name="Loop", layer_name="loop"),
    "Rwd": TransportButton(control_name="Rwd", layer_name="seek_backward"),
    "Ffwd": TransportButton(control_name="FFwd", layer_name="seek_forward"),
    "PunchIn": TransportButton(control_name="Punch_In", layer_name="punch_in"),
    "PunchOut": TransportButton(control_name="Punch_Out", layer_name="punch_out"),
    "NudgeUp": TransportButton(control_name="Nudge_Up", layer_name="nudge_up"),
    "NudgeDown": TransportButton(control_name="Nudge_Down", layer_name="nudge_down"),
    "TapTempo": TransportButton(control_name="Tap_Tempo", layer_name="tap_tempo"),
}


class GenericScript(ControlSurface):
    def __init__(
        self,
        c_instance,
        macro_map_mode,
        volume_map_mode,
        device_controls,
        transport_controls,
        volume_controls,
        trackarm_controls,
        bank_controls,
        descriptions=None,
        mixer_options=None,
        device_component_class=DeviceComponent,
    ):
        ControlSurface.__init__(self, c_instance)
        with self.component_guard():
            global_channel = 0
            if descriptions:
                if "INPUTPORT" in descriptions:
                    self._suggested_input_port = descriptions["INPUTPORT"]
                if "OUTPUTPORT" in descriptions:
                    self._suggested_output_port = descriptions["OUTPUTPORT"]
                if "CHANNEL" in descriptions:
                    if is_valid_midi_channel(descriptions["CHANNEL"]):
                        global_channel = descriptions["CHANNEL"]
                if "PAD_TRANSLATION" in descriptions:
                    self.set_pad_translations(descriptions["PAD_TRANSLATION"])
            self._init_mixer_component(
                volume_controls,
                trackarm_controls,
                mixer_options,
                global_channel,
                volume_map_mode,
            )
            self._init_device_component(
                device_controls,
                bank_controls,
                global_channel,
                macro_map_mode,
                device_component_class,
            )
            self._init_transport_component(transport_controls, global_channel)

    def handle_sysex(self, midi_bytes):
        pass

    def _init_mixer_component(self, volume_controls, trackarm_controls, mixer_options, global_channel, volume_map_mode):
        momentary_buttons = mixer_options is not None and "NOTOGGLE" in list(mixer_options.keys())
        sends_map_mode = Live.MidiMap.MapMode.absolute
        if mixer_options:
            if "SENDMAPMODE" in mixer_options:
                sends_map_mode = mixer_options["SENDMAPMODE"]
        MixerButton = partial(ButtonElement, momentary_buttons, MIDI_CC_TYPE, global_channel)

        def make_mixer_encoder(cc, channel, name, map_mode=volume_map_mode):
            if is_valid_midi_identifier(cc):
                if is_valid_midi_channel(channel):
                    return EncoderElement(MIDI_CC_TYPE, channel, cc, map_mode, name=name)
                return None
            return None

        def make_global_mixer_encoder(identifier_spec, channel_spec, control_name, **k):
            if has_specification_for(identifier_spec, mixer_options):
                channel = global_channel
                if channel_spec in mixer_options:
                    if is_valid_midi_channel(mixer_options[channel_spec]):
                        channel = mixer_options[channel_spec]
                return make_mixer_encoder((mixer_options[identifier_spec]), channel, control_name, **k)
            return None

        def make_mixer_button(control, name):
            return MixerButton((mixer_options[control]), name=name)

        def make_channel_strip_button(control_list, index, name):
            if 0 <= index < len(control_list):
                if is_valid_midi_identifier(control_list[index]):
                    return MixerButton((control_list[index]), name=(f"{index}_{name}_Button"))
                return None
            return None

        if volume_controls is not None:
            if trackarm_controls is not None:
                num_strips = max(len(volume_controls), len(trackarm_controls))
                send_info = []
                mute_controls = []
                solo_controls = []
                select_controls = []
                mixer = SpecialMixerComponent(
                    num_strips,
                    name="Mixer",
                    invert_mute_feedback=(mixer_options and mixer_options.get("INVERTMUTELEDS", True)),
                )
                mixer.master_strip().name = "Master_Channel_Strip"
                mixer.selected_strip().name = "Selected_Channel_Strip"
                if mixer_options is not None:
                    master_volume = make_global_mixer_encoder(
                        "MASTERVOLUME",
                        "MASTERVOLUMECHANNEL",
                        "Master_Volume_Control",
                    )
                    if master_volume:
                        mixer.master_strip().layer = Layer(volume_control=master_volume)
                    for send in range(mixer_options.get("NUMSENDS", 0)):
                        send_info.append(mixer_options[f"SEND{send + 1}"])

                    layer_specs = {}
                    cue_volume = make_global_mixer_encoder("CUEVOLUME", "CUEVOLUMECHANNEL", "Cue_Volume_Control")
                    if cue_volume:
                        layer_specs["prehear_volume_control"] = cue_volume
                    crossfader_map_mode = Live.MidiMap.MapMode.absolute
                    if "CROSSFADERMAPMODE" in mixer_options:
                        crossfader_map_mode = mixer_options["CROSSFADERMAPMODE"]
                    crossfader = make_global_mixer_encoder(
                        "CROSSFADER",
                        "CROSSFADERCHANNEL",
                        "Crossfader_Control",
                        map_mode=crossfader_map_mode,
                    )
                    if crossfader:
                        layer_specs["crossfader_control"] = crossfader
                        crossfader.set_needs_takeover(False)
                    if has_specification_for("NEXTBANK", mixer_options):
                        layer_specs["bank_up_button"] = make_mixer_button("NEXTBANK", "Mixer_Next_Bank_Button")
                    if has_specification_for("PREVBANK", mixer_options):
                        layer_specs["bank_down_button"] = make_mixer_button("PREVBANK", "Mixer_Previous_Bank_Button")
                    mute_controls = mixer_options.get("MUTE", [])
                    solo_controls = mixer_options.get("SOLO", [])
                    select_controls = mixer_options.get("SELECT", [])
                    mixer.layer = Layer(**layer_specs)
                for track in range(num_strips):
                    strip = mixer.channel_strip(track)
                    strip.name = f"Channel_Strip_{track}"
                    layer_specs = {}
                    if 0 <= track < len(volume_controls):
                        channel = global_channel
                        cc = volume_controls[track]
                        if isinstance(volume_controls[track], (tuple, list)):
                            cc = volume_controls[track][0]
                            if is_valid_midi_channel(volume_controls[track][1]):
                                channel = volume_controls[track][1]
                        if is_valid_midi_identifier(cc):
                            if is_valid_midi_channel(channel):
                                layer_specs["volume_control"] = make_mixer_encoder(
                                    cc,
                                    channel,
                                    f"{track}_Volume_Control",
                                )
                    arm_button = make_channel_strip_button(trackarm_controls, track, "Arm")
                    if arm_button:
                        layer_specs["arm_button"] = arm_button
                    else:
                        mute_button = make_channel_strip_button(mute_controls, track, "Mute")
                        if mute_button:
                            layer_specs["mute_button"] = mute_button
                        solo_button = make_channel_strip_button(solo_controls, track, "Solo")
                        if solo_button:
                            layer_specs["solo_button"] = solo_button
                        select_button = make_channel_strip_button(select_controls, track, "Select")
                        if select_button:
                            layer_specs["select_button"] = select_button
                        send_controls_raw = []
                        for index, send in enumerate(send_info):
                            if 0 <= track < len(send):
                                channel = global_channel
                                cc = send[track]
                                if isinstance(send[track], (tuple, list)):
                                    cc = send[track][0]
                                    if is_valid_midi_channel(send[track][1]):
                                        channel = send[track][1]
                                if is_valid_midi_identifier(cc):
                                    if is_valid_midi_channel(channel):
                                        send_controls_raw.append(
                                            make_mixer_encoder(
                                                cc,
                                                channel,
                                                name=(f"{track}_Send_{index}_Control"),
                                                map_mode=sends_map_mode,
                                            ),
                                        )

                        if len(send_controls_raw) > 0:
                            layer_specs["send_controls"] = ButtonMatrixElement(
                                rows=[send_controls_raw],
                                name=(f"{track}_Send_Controls"),
                            )
                        strip.layer = Layer(**layer_specs)

    def _init_device_component(
        self,
        device_controls,
        bank_controls,
        global_channel,
        macro_map_mode,
        device_component_class,
    ):
        is_momentary = True
        DeviceButton = partial(ButtonElement, is_momentary, MIDI_CC_TYPE)

        def make_bank_button(control, name):
            return DeviceButton(global_channel, (bank_controls[control]), name=name)

        if device_controls:
            layer_specs = {}
            if bank_controls:
                if has_specification_for("NEXTBANK", bank_controls):
                    layer_specs["bank_next_button"] = make_bank_button("NEXTBANK", "Device_Next_Bank_Button")
                if has_specification_for("PREVBANK", bank_controls):
                    layer_specs["bank_prev_button"] = make_bank_button("PREVBANK", "Device_Previous_Bank_Button")
                if has_specification_for("TOGGLELOCK", bank_controls):
                    layer_specs["lock_button"] = make_bank_button("TOGGLELOCK", "Device_Lock_Button")
                if has_specification_for("ONOFF", bank_controls):
                    layer_specs["on_off_button"] = make_bank_button("ONOFF", "Device_On_Off_Button")
                bank_buttons_raw = []
                for index in range(8):
                    key = f"BANK{index + 1}"
                    if key in list(bank_controls.keys()):
                        control_info = bank_controls[key]
                        channel = global_channel
                        cc = control_info
                        if isinstance(control_info, (tuple, list)):
                            cc = control_info[0]
                            if is_valid_midi_channel(control_info[1]):
                                channel = control_info[1]
                        if is_valid_midi_identifier(cc):
                            if is_valid_midi_channel(channel):
                                name = f"Device_Bank_{index}_Button"
                                bank_buttons_raw.append(DeviceButton(channel, cc, name=name))

                if len(bank_buttons_raw) > 0:
                    layer_specs["bank_buttons"] = ButtonMatrixElement(
                        rows=[bank_buttons_raw],
                        name="Device_Bank_Buttons",
                    )
            parameter_encoders_raw = []
            for index, control_info in enumerate(device_controls):
                channel = global_channel
                cc = control_info
                if isinstance(control_info, (tuple, list)):
                    cc = control_info[0]
                    if is_valid_midi_channel(control_info[1]):
                        channel = control_info[1]
                if is_valid_midi_identifier(cc):
                    if is_valid_midi_channel(channel):
                        name = "Device_Parameter_%d_Control" % index
                        parameter_encoders_raw.append(
                            EncoderElement(MIDI_CC_TYPE, channel, cc, macro_map_mode, name=name),
                        )

            if len(parameter_encoders_raw) > 0:
                layer_specs["parameter_controls"] = ButtonMatrixElement(
                    rows=[parameter_encoders_raw],
                    name="Device_Parameter_Controls",
                )
            if layer_specs:
                device = device_component_class(device_selection_follows_track_selection=True, name="Device_Component")
                device.layer = Layer(**layer_specs)
                self.set_device_component(device)

    def _init_transport_component(self, transport_controls, global_channel):
        def make_transport_button(control, name, is_momentary=True):
            return ButtonElement(is_momentary, MIDI_CC_TYPE, global_channel, (transport_controls[control]), name=name)

        if transport_controls:
            momentary_seek = "NORELEASE" not in list(transport_controls.keys())
            layer_specs = {}
            for key, spec in iteritems(TRANSPORT_BUTTON_SPECIFICATIONS):
                key_upper = key.upper()
                if has_specification_for(key_upper, transport_controls):
                    layer_specs[f"{spec.layer_name}_button"] = make_transport_button(
                        key_upper,
                        (f"{spec.control_name}_Button"),
                        is_momentary=(momentary_seek if key in ("Rwd", "Ffwd") else True),
                    )

            transport = TransportComponent(name="Transport")
            transport.layer = Layer(**layer_specs)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Generic/SelectChanStripComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1046 bytes
from _Framework import ChannelStripComponent


class SelectChanStripComponent(ChannelStripComponent):
    def __init__(self):
        ChannelStripComponent.__init__(self)

    def _arm_value(self, value):
        if self.is_enabled():
            track_was_armed = False
            if self._track is not None:
                if self._track.can_be_armed:
                    track_was_armed = self._track.arm
            ChannelStripComponent._arm_value(self, value)
            if self._track is None and self._track.can_be_armed or self._track.arm:
                if not track_was_armed:
                    if self._track.view.select_instrument():
                        self.song().view.selected_track = self._track



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/control/text_display.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1485 bytes
from ..elements import DisplayDataSource
from .control import Control


class TextDisplayControl(Control):
    class State(Control.State):
        def __init__(self, segments=("",), *a, **k):
            (super(TextDisplayControl.State, self).__init__)(*a, **k)
            self._data_sources = [DisplayDataSource(segment) for segment in segments]

        def set_control_element(self, control_element):
            if not control_element:
                if self._control_element:
                    self._control_element.set_data_sources(None)
            super(TextDisplayControl.State, self).set_control_element(control_element)
            if control_element:
                control_element.set_data_sources(self._data_sources)

        def __getitem__(self, index):
            return self._data_sources[index].display_string()

        def __setitem__(self, index, value):
            return self._data_sources[index].set_display_string(value)

    def __init__(self, *a, **k):
        super().__init__(extra_args=a, extra_kws=k)


class ConfigurableTextDisplayControl(TextDisplayControl):
    class State(TextDisplayControl.State):
        def set_data_sources(self, data_sources):
            self._data_sources = data_sources
            if self._control_element:
                self._control_element.set_data_sources(data_sources)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/ClipSlotComponent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 10398 bytes
import Live

from .ControlSurfaceComponent import ControlSurfaceComponent
from .SubjectSlot import subject_slot
from .Util import in_range


def find_nearest_color(rgb_table, src_hex_color):
    def hex_to_channels(color_in_hex):
        return ((color_in_hex & 16711680) >> 16, (color_in_hex & 65280) >> 8, color_in_hex & 255)

    def squared_distance(color):
        return sum([(a - b) ** 2 for a, b in zip(hex_to_channels(src_hex_color), hex_to_channels(color[1]))])

    return min(rgb_table, key=squared_distance)[0]


class ClipSlotComponent(ControlSurfaceComponent):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._clip_slot = None
        self._triggered_to_play_value = 126
        self._triggered_to_record_value = 121
        self._started_value = 127
        self._recording_value = 120
        self._stopped_value = 0
        self._clip_palette = []
        self._clip_rgb_table = None
        self._record_button_value = None
        self._has_fired_slot = False
        self._delete_button = None
        self._select_button = None
        self._duplicate_button = None

    def on_enabled_changed(self):
        self.update()

    def set_clip_slot(self, clip_slot):
        self._clip_slot = clip_slot
        self._update_clip_property_slots()
        self._on_slot_triggered_changed.subject = clip_slot
        self._on_slot_playing_state_changed.subject = clip_slot
        self._on_clip_state_changed.subject = clip_slot
        self._on_controls_other_clips_changed.subject = clip_slot
        self._on_has_stop_button_changed.subject = clip_slot
        self._on_clip_slot_color_changed.subject = clip_slot
        track = clip_slot.canonical_parent if clip_slot else None
        if track:
            if track in self.song().tracks:
                self._on_arm_value_changed.subject = track
                self._on_implicit_arm_value_changed.subject = track
                self._on_input_routing_type_changed.subject = track
        self.update()

    def set_launch_button(self, button):
        self._launch_button_value.subject = button
        self.update()

    def set_delete_button(self, button):
        self._delete_button = button

    def set_select_button(self, button):
        self._select_button = button

    def set_duplicate_button(self, button):
        self._duplicate_button = button

    def set_triggered_to_play_value(self, value):
        self._triggered_to_play_value = value

    def set_triggered_to_record_value(self, value):
        self._triggered_to_record_value = value

    def set_started_value(self, value):
        self._started_value = value

    def set_recording_value(self, value):
        self._recording_value = value

    def set_stopped_value(self, value):
        self._stopped_value = value
        self._clip_palette = []

    def set_record_button_value(self, value):
        self._record_button_value = value

    def set_clip_palette(self, palette):
        self._stopped_value = None
        self._clip_palette = palette

    def set_clip_rgb_table(self, rgb_table):
        self._clip_rgb_table = rgb_table

    def has_clip(self):
        return self._clip_slot.has_clip

    def update(self):
        super().update()
        self._has_fired_slot = False
        button = self._launch_button_value.subject
        if self._allow_updates:
            if self.is_enabled():
                if button is not None:
                    value_to_send = self._feedback_value()
                    if value_to_send in (None, -1):
                        button.turn_off()
                else:
                    pass
                if in_range(value_to_send, 0, 128):
                    button.send_value(value_to_send)
                else:
                    button.set_light(value_to_send)
        else:
            self._update_requests += 1

    def _color_value(self, color):
        try:
            return self._clip_palette[color]
        except (KeyError, IndexError):
            if self._clip_rgb_table is not None:
                return find_nearest_color(self._clip_rgb_table, color)
            else:
                return self._stopped_value

    def _track_is_armed(self, track):
        return track is not None and track.can_be_armed and any([track.arm, track.implicit_arm])

    def _feedback_value(self):
        if self._clip_slot is not None:
            track = self._clip_slot.canonical_parent
            slot_or_clip = self._clip_slot.clip if self.has_clip() else self._clip_slot
            if slot_or_clip.is_triggered:
                if slot_or_clip.will_record_on_start:
                    return self._triggered_to_record_value
                return self._triggered_to_play_value
            if slot_or_clip.is_playing:
                if slot_or_clip.is_recording:
                    return self._recording_value
                return self._started_value
            if slot_or_clip.color is not None:
                return self._color_value(slot_or_clip.color)
            if getattr(slot_or_clip, "controls_other_clips", True):
                if self._stopped_value is not None:
                    return self._stopped_value
            if self._track_is_armed(track):
                if self._clip_slot.has_stop_button:
                    if self._record_button_value is not None:
                        return self._record_button_value
                    return None
                return None
            return None
        return None

    def _update_clip_property_slots(self):
        clip = self._clip_slot.clip if self._clip_slot else None
        self._on_clip_playing_state_changed.subject = clip
        self._on_recording_state_changed.subject = clip
        self._on_clip_color_changed.subject = clip

    @subject_slot("has_clip")
    def _on_clip_state_changed(self):
        self._update_clip_property_slots()
        self.update()

    @subject_slot("controls_other_clips")
    def _on_controls_other_clips_changed(self):
        self._update_clip_property_slots()
        self.update()

    @subject_slot("color")
    def _on_clip_color_changed(self):
        self.update()

    @subject_slot("color")
    def _on_clip_slot_color_changed(self):
        self.update()

    @subject_slot("playing_status")
    def _on_slot_playing_state_changed(self):
        self.update()

    @subject_slot("playing_status")
    def _on_clip_playing_state_changed(self):
        self.update()

    @subject_slot("is_recording")
    def _on_recording_state_changed(self):
        self.update()

    @subject_slot("arm")
    def _on_arm_value_changed(self):
        self.update()

    @subject_slot("implicit_arm")
    def _on_implicit_arm_value_changed(self):
        self.update()

    @subject_slot("input_routing_type")
    def _on_input_routing_type_changed(self):
        self.update()

    @subject_slot("has_stop_button")
    def _on_has_stop_button_changed(self):
        self.update()

    @subject_slot("is_triggered")
    def _on_slot_triggered_changed(self):
        if not self.has_clip():
            song = self.song()
            view = song.view
            if song.select_on_launch:
                if self._clip_slot.is_triggered:
                    if self._has_fired_slot:
                        if self._clip_slot.will_record_on_start:
                            if self._clip_slot != view.highlighted_clip_slot:
                                view.highlighted_clip_slot = self._clip_slot
            self.update()

    @subject_slot("value")
    def _launch_button_value(self, value):
        if self.is_enabled():
            if self._select_button and self._select_button.is_pressed() and value:
                self._do_select_clip(self._clip_slot)
            else:
                if self._clip_slot is None and self._duplicate_button and self._duplicate_button or value:
                    self._do_duplicate_clip()
                else:
                    if self._delete_button and self._delete_button.is_pressed():
                        if value:
                            self._do_delete_clip()
                    else:
                        self._do_launch_clip(value)

    def _do_delete_clip(self):
        if self._clip_slot:
            if self._clip_slot.has_clip:
                self._clip_slot.delete_clip()

    def _do_select_clip(self, clip_slot):
        if self._clip_slot is not None:
            if self.song().view.highlighted_clip_slot != self._clip_slot:
                self.song().view.highlighted_clip_slot = self._clip_slot

    def _do_duplicate_clip(self):
        if not self._clip_slot or self._clip_slot.has_clip:
            try:
                track = self._clip_slot.canonical_parent
                track.duplicate_clip_slot(list(track.clip_slots).index(self._clip_slot))
            except Live.Base.LimitationError:
                pass
            except RuntimeError:
                pass

    def _do_launch_clip(self, value):
        button = self._launch_button_value.subject
        object_to_launch = self._clip_slot
        launch_pressed = value or not button.is_momentary()
        if self.has_clip():
            object_to_launch = self._clip_slot.clip
        else:
            self._has_fired_slot = True
        if button.is_momentary():
            object_to_launch.set_fire_button_state(value != 0)
        else:
            if launch_pressed:
                object_to_launch.fire()
        if launch_pressed:
            if self.has_clip():
                if self.song().select_on_launch:
                    self.song().view.highlighted_clip_slot = self._clip_slot



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/session_ring_selection_linking.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2544 bytes
from base.dependency import depends
from base.event import EventObject, listens
from base.util import index_if


class SessionRingSelectionLinking(EventObject):
    @depends(song=None)
    def __init__(self, session_ring=None, selection_changed_notifier=None, song=None, *a, **k):
        (super().__init__)(*a, **k)
        self._session_ring = session_ring
        self._song = song
        self._on_selection_scrolled.subject = selection_changed_notifier

    @listens("selection_scrolled")
    def _on_selection_scrolled(self):
        self._link_session_ring_with_minimal_travel()

    def _link_session_ring_with_minimal_travel(self):
        if self._song.view.selected_track == self._song.master_track:
            return
        track_index = self._current_track_index()
        right_ring_index = self._session_ring.track_offset + self._session_ring.num_tracks - 1
        offset_left = track_index - self._session_ring.track_offset
        offset_right = track_index - right_ring_index
        adjustment = min(0, offset_left) + max(0, offset_right)
        new_track_offset = self._session_ring.track_offset + adjustment
        if new_track_offset != self._session_ring.track_offset:
            self._session_ring.set_offsets(new_track_offset, self._session_ring.scene_offset)

    def _current_track(self):
        return self._song.view.selected_track

    def _current_track_index(self):
        return index_if(lambda t: t == self._current_track(), self._session_ring.tracks_to_use())



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_Framework/SliderElement.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 619 bytes
import Live

from .EncoderElement import EncoderElement


class SliderElement(EncoderElement):
    def __init__(self, msg_type, channel, identifier, *a, **k):
        (super().__init__)(msg_type, channel, identifier, *a, map_mode=Live.MidiMap.MapMode.absolute, **k)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/elements/playhead_element.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 681 bytes
from ...base import nop
from .proxy_element import ProxyElement


class NullPlayhead:
    notes = []
    start_time = 0.0
    step_length = 1.0
    velocity = 0.0
    wrap_around = False
    track = None
    clip = None
    set_feedback_channels = nop


class PlayheadElement(ProxyElement):
    def __init__(self, playhead=None, *a, **k):
        super().__init__(proxied_object=playhead, proxied_interface=(NullPlayhead()))

    def reset(self):
        self.track = None



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/device_bank_navigation.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 6944 bytes
from math import ceil
from typing import cast

from ...base import depends, listens
from ..controls import FixedRadioButtonGroup
from ..display import Renderable
from .scroll import Scrollable, ScrollComponent

NUM_BANK_SELECT_BUTTONS = 8


class DeviceBankNavigationComponent(ScrollComponent, Renderable, Scrollable):
    bank_select_buttons = FixedRadioButtonGroup(
        control_count=NUM_BANK_SELECT_BUTTONS,
        unchecked_color="Device.Bank.NotSelected",
        checked_color="Device.Bank.Selected",
    )

    @depends(device_bank_registry=None)
    def __init__(self, name="Device_Bank_Navigation", banking_info=None, device_bank_registry=None, *a, **k):
        (super().__init__)(a, name=name, scroll_skin_name="Device.Bank.Navigation", **k)
        self._bank_provider = None
        self._banking_info = banking_info
        self._device_bank_registry = device_bank_registry

    @property
    def bank_provider(self):
        return self._bank_provider

    @bank_provider.setter
    def bank_provider(self, provider):
        self._bank_provider = provider
        self._sync_registry()
        self._DeviceBankNavigationComponent__on_provider_bank_changed.subject = provider
        self._DeviceBankNavigationComponent__on_parameters_changed_in_device.subject = (
            provider.device if provider else None
        )
        self.update()

    def set_bank_scroll_encoder(self, encoder):
        self.set_scroll_encoder(encoder)

    def set_prev_bank_button(self, button):
        self.set_scroll_down_button(button)

    def set_next_bank_button(self, button):
        self.set_scroll_up_button(button)

    def set_bank_select_buttons(self, buttons):
        self.bank_select_buttons.set_control_element(buttons)
        self._sync_registry()

    def can_scroll_up(self):
        return self._bank_provider is not None and (
            self.scroll_down_button.control_element is None
            or self._bank_provider.index < self._bank_provider.bank_count() - self._banking_info.num_simultaneous_banks
        )

    def can_scroll_down(self):
        return self._bank_provider is not None and self._bank_provider.index > 0

    def scroll_up(self):
        roundtrip_banking = self.scroll_down_button.control_element is None and self.scroll_up_button.is_pressed
        new_index = self._bank_provider.index + self._banking_info.num_simultaneous_banks
        if roundtrip_banking:
            if new_index >= self._bank_provider.bank_count():
                new_index = 0
        self._bank_provider.index = new_index
        self._notify_bank_name()

    def scroll_down(self):
        self._bank_provider.index = self._bank_provider.index - self._banking_info.num_simultaneous_banks
        self._notify_bank_name()

    @bank_select_buttons.checked
    def bank_select_buttons(self, button):
        self._bank_provider.index = button.index * self._banking_info.num_simultaneous_banks + int(
            self._should_skip_first_bank(),
        )
        self._notify_bank_name()

    def _notify_bank_name(self):
        self.notify(
            self.notifications.Device.bank,
            cast(str, self._banking_info.device_bank_names(self._bank_provider.device)[self._bank_provider.index]),
        )

    def _adjusted_bank_count(self):
        if self._bank_provider:
            return ceil(self._bank_provider.bank_count() / self._banking_info.num_simultaneous_banks)
        return 0

    def _should_skip_first_bank(self):
        return (
            self.bank_select_buttons[0].control_element is not None
            and self._bank_provider is not None
            and self._banking_info.has_main_bank(self._bank_provider.device)
            and self._adjusted_bank_count() > NUM_BANK_SELECT_BUTTONS
        )

    def _sync_registry(self):
        registry = self._device_bank_registry
        device = self._bank_provider.device if self._bank_provider is not None else None
        if self._should_skip_first_bank():
            if registry.get_device_bank(device) == 0:
                registry.set_device_bank(device, 1)

    @listens("parameters")
    def __on_provider_bank_changed(self):
        self.update()

    @listens("parameters")
    def __on_parameters_changed_in_device(self):
        self.update()

    def update(self):
        super().update()
        self._update_bank_select_buttons()

    def _update_bank_select_buttons(self):
        bank_index = ceil(
            (self._bank_provider.index if self._bank_provider else 0) / self._banking_info.num_simultaneous_banks,
        )
        bank_count = self._adjusted_bank_count()
        if self._should_skip_first_bank():
            bank_count -= 1
            bank_index -= 1
        has_banks = bank_count > 1
        self.bank_select_buttons.active_control_count = bank_count if has_banks else 0
        if has_banks:
            self.bank_select_buttons.checked_index = bank_index if bank_index < NUM_BANK_SELECT_BUTTONS else -1



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/mode/modes.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 12552 bytes
from functools import partial
from typing import cast

from ableton.v2.control_surface import StackingResource
from ableton.v2.control_surface.mode import _ModeEntry, tomode

from ...base import listenable_property, task
from .. import Component
from ..controls import ButtonControl, SendValueInputControl
from ..display import Renderable
from . import ImmediateBehaviour, make_mode_button_control


class ModesComponent(Component, Renderable):
    mode_selection_control = SendValueInputControl()
    cycle_mode_button = ButtonControl()
    default_behaviour = ImmediateBehaviour()
    previous_mode = listenable_property.managed(None)

    def __init__(
        self,
        name=None,
        support_momentary_mode_cycling=True,
        default_behaviour=None,
        is_private=False,
        *a,
        **k,
    ):
        (super().__init__)(a, name=name, is_private=is_private, **k)
        self._support_momentary_mode_cycling = support_momentary_mode_cycling
        self._mode_list = []
        self._mode_map = {}
        self._last_selected_mode = None
        self._mode_stack = StackingResource(self._do_enter_mode, self._do_leave_mode)
        self._push_mode_tasks = {}
        self._pop_mode_tasks = {}
        if default_behaviour is not None:
            self.default_behaviour = default_behaviour

    def disconnect(self):
        self._mode_stack.release_all()
        super().disconnect()

    @property
    def modes(self):
        return self._mode_list

    @property
    def active_modes(self):
        return self._mode_stack.clients

    @listenable_property
    def selected_mode(self):
        return self._mode_stack.owner or self._last_selected_mode

    @selected_mode.setter
    def selected_mode(self, mode):
        self.previous_mode = self.selected_mode
        if self.is_enabled():
            if self.selected_mode != mode:
                if mode is not None:
                    self.push_mode(mode)
                    self.pop_unselected_modes()
                    self.notify(self.notifications.Modes.select, cast(str, self.name), cast(str, mode))
                else:
                    self._mode_stack.release_all()
        else:
            self._last_selected_mode = mode

    def get_mode(self, name):
        entry = self._mode_map.get(name, None)
        return entry and entry.mode

    def get_mode_button(self, name):
        return getattr(self, f"{name}_button")

    def get_mode_groups(self, name):
        entry = self._mode_map.get(name, None)
        if entry:
            return entry.groups
        return set()

    def push_mode(self, mode, delay=0):
        self._cancel_push_mode_task(mode)
        if mode in self._pop_mode_tasks:
            self._cancel_pop_mode_task(mode)
        else:
            if not delay:
                self._do_push_mode(mode)
            else:
                self._push_mode_tasks[mode] = self._tasks.add(
                    task.sequence(task.wait(delay), task.run(partial(self._do_push_mode, mode))),
                )

    def _do_push_mode(self, mode):
        self._cancel_push_mode_task(mode)
        self._mode_stack.grab(mode)

    def _cancel_push_mode_task(self, mode):
        if mode in self._push_mode_tasks:
            self._push_mode_tasks[mode].kill()
            self._tasks.remove(self._push_mode_tasks[mode])
            del self._push_mode_tasks[mode]

    def pop_mode(self, mode, delay=0):
        self._cancel_pop_mode_task(mode)
        if mode in self._push_mode_tasks:
            self._cancel_push_mode_task(mode)
        else:
            if not delay:
                self._do_pop_mode(mode)
            else:
                self._pop_mode_tasks[mode] = self._tasks.add(
                    task.sequence(task.wait(delay), task.run(partial(self._do_pop_mode, mode))),
                )

    def _do_pop_mode(self, mode):
        self._cancel_pop_mode_task(mode)
        if len(self.active_modes) <= 1:
            return
        self._mode_stack.release(mode)

    def _cancel_pop_mode_task(self, mode):
        if mode in self._pop_mode_tasks:
            self._pop_mode_tasks[mode].kill()
            self._tasks.remove(self._pop_mode_tasks[mode])
            del self._pop_mode_tasks[mode]

    def pop_unselected_modes(self):
        self._mode_stack.release_stacked()

    def pop_groups(self, groups):
        if not isinstance(groups, set):
            groups = set(groups)
        for client in self._mode_stack.clients:
            if self.get_mode_groups(client) & groups:
                self._mode_stack.release(client)

    def cycle_mode(self, delta=1):
        current_index = self._mode_list.index(self.selected_mode) if self.selected_mode else -delta
        current_index = (current_index + delta) % len(self._mode_list)
        self.selected_mode = self._mode_list[current_index]

    def add_mode(self, name, mode_or_component, groups=None, behaviour=None, selector=None):
        if not isinstance(groups, set):
            groups = set(groups) if groups is not None else set()
        mode = tomode(mode_or_component)
        behaviour = behaviour or self.default_behaviour
        self._mode_list.append(name)
        self._mode_map[name] = _ModeEntry(
            mode=mode,
            cycle_mode_button_color=(f"{self._get_mode_color_base_name(name)}.On"),
            behaviour=behaviour,
            groups=groups,
        )
        self.add_mode_button_control(name, behaviour)
        if callable(selector):
            selector(self, name)

    def add_mode_button_control(self, mode_name, behaviour):
        mode_color_basename = self._get_mode_color_base_name(mode_name)
        colors = {
            "mode_selected_color": f"{mode_color_basename}.On",
            "mode_unselected_color": f"{mode_color_basename}.Off",
            "mode_group_active_color": f"{mode_color_basename}.On",
        }
        button_control = make_mode_button_control(self, mode_name, behaviour, **colors)
        self.add_control(f"{mode_name}_button", button_control)
        self._update_mode_buttons(self.selected_mode)

    @mode_selection_control.value
    def mode_selection_control(self, value, _):
        modes = self.modes
        if value < len(modes):
            self.selected_mode = modes[value]

    @cycle_mode_button.pressed
    def cycle_mode_button(self, _):
        if self._mode_list:
            self.cycle_mode(1)

    @cycle_mode_button.released_delayed
    def cycle_mode_button(self, _):
        if self._mode_list:
            if self._support_momentary_mode_cycling:
                self.cycle_mode(-1)

    def _do_enter_mode(self, name):
        entry = self._mode_map[name]
        entry.mode.enter_mode()
        self._update_mode_buttons(name)
        self._update_cycle_mode_button(name)
        self.notify_selected_mode(name)

    def _do_leave_mode(self, name):
        self._mode_map[name].mode.leave_mode()
        if self._mode_stack.stack_size == 0:
            self._update_mode_buttons(None)
            self._update_cycle_mode_button(None)
            self.notify_selected_mode(None)

    def _get_mode_behaviour(self, name):
        entry = self._mode_map.get(name, None)
        if entry is not None:
            return entry.behaviour
        return self.default_behaviour

    def _get_mode_color_base_name(self, mode_name):
        return "{}.{}".format(self.name.title().replace("_", ""), mode_name.title().replace("_", ""))

    def on_enabled_changed(self):
        super().on_enabled_changed()
        if not self.is_enabled():
            self._last_selected_mode = self.selected_mode
            self._mode_stack.release_all()
        else:
            if self._last_selected_mode:
                self.push_mode(self._last_selected_mode)

    def _update_mode_buttons(self, selected):
        if self.is_enabled():
            for name in self._mode_map:
                self._get_mode_behaviour(name).update_button(self, name, selected)

        if selected in self._mode_list:
            self.mode_selection_control.value = self._mode_list.index(selected)

    def _update_cycle_mode_button(self, selected):
        entry = self._mode_map.get(selected)
        color = entry.cycle_mode_button_color if entry else None
        if color is not None:
            self.cycle_mode_button.color = color



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/mode/behaviour.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3733 bytes
from ...base import mixin
from . import ModeButtonBehaviour, pop_last_mode


def make_reenter_behaviour(base_behaviour, on_reenter=None, *a, **k):
    return (mixin(ReenterBehaviourMixin, base_behaviour))(a, on_reenter=on_reenter, **k)


class ReenterBehaviourMixin:
    def __init__(self, on_reenter=None, *a, **k):
        (super().__init__)(*a, **k)
        self._on_reenter = on_reenter

    def press_immediate(self, component, mode):
        was_active = component.selected_mode == mode
        super().press_immediate(component, mode)
        if was_active:
            self._on_reenter()


class ToggleBehaviour(ModeButtonBehaviour):
    def __init__(self, return_to_default=False, *a, **k):
        (super().__init__)(*a, **k)
        self._return_to_default = return_to_default

    def press_immediate(self, component, mode):
        if component.selected_mode == mode:
            if self._return_to_default:
                component.push_mode(component.modes[0])
                component.pop_unselected_modes()
            else:
                pop_last_mode(component, mode)
        else:
            component.push_mode(mode)


class MomentaryBehaviour(ModeButtonBehaviour):
    def __init__(self, entry_delay=None, exit_delay=None, immediate_exit_delay=None):
        self._entry_delay = entry_delay
        self._exit_delay = exit_delay
        self._immediate_exit_delay = self._exit_delay if immediate_exit_delay is None else immediate_exit_delay

    def press_immediate(self, component, mode):
        component.push_mode(mode, delay=(self._entry_delay))

    def release_immediate(self, component, mode):
        component.pop_mode(mode, delay=(self._immediate_exit_delay))

    def release_delayed(self, component, mode):
        component.pop_mode(mode, delay=(self._exit_delay))



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/control/button.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 15001 bytes
from functools import partial

from ...base import lazy_attribute, task
from ..defaults import DOUBLE_CLICK_DELAY, MOMENTARY_DELAY
from ..input_control_element import ScriptForwarding
from .control import InputControl, control_color, control_event

__all__ = ("ButtonControl", "PlayableControl", "ButtonControlBase", "DoubleClickContext")


class DoubleClickContext:
    control_state = None
    click_count = 0

    def set_new_context(self, control_state):
        self.control_state = control_state
        self.click_count = 0


class ButtonControlBase(InputControl):
    DELAY_TIME = MOMENTARY_DELAY
    DOUBLE_CLICK_TIME = DOUBLE_CLICK_DELAY
    REPEAT_RATE = 0.1
    pressed = control_event("pressed")
    released = control_event("released")
    pressed_delayed = control_event("pressed_delayed")
    released_delayed = control_event("released_delayed")
    released_immediately = control_event("released_immediately")
    double_clicked = control_event("double_clicked")

    class State(InputControl.State):
        disabled_color = control_color("DefaultButton.Disabled")
        pressed_color = control_color(None)

        def __init__(
            self,
            pressed_color=None,
            disabled_color=None,
            repeat=False,
            enabled=True,
            double_click_context=None,
            delay_time=None,
            *a,
            **k,
        ):
            (super(ButtonControlBase.State, self).__init__)(*a, **k)
            if disabled_color is not None:
                self.disabled_color = disabled_color
            self.pressed_color = pressed_color
            self._repeat = repeat
            self._is_pressed = False
            self._enabled = enabled
            self._double_click_context = double_click_context or DoubleClickContext()
            self._delay_time = delay_time if delay_time is not None else ButtonControlBase.DELAY_TIME

        @property
        def enabled(self):
            return self._enabled

        @enabled.setter
        def enabled(self, enabled):
            if self._enabled != enabled:
                if not enabled:
                    self._release_button()
                self._enabled = enabled
                self._send_current_color()

        @property
        def is_momentary(self):
            return self._control_element and self._control_element.is_momentary()

        @property
        def is_pressed(self):
            return self._is_pressed

        def _event_listener_required(self):
            return True

        def set_control_element(self, control_element):
            if self._control_element != control_element:
                self._release_button()
                self._kill_all_tasks()
            super(ButtonControlBase.State, self).set_control_element(control_element)
            self._send_current_color()

        def _send_current_color(self):
            if self._control_element:
                if not self._enabled:
                    self._control_element.set_light(self.disabled_color)
                else:
                    if self.pressed_color is not None and self.is_pressed:
                        self._control_element.set_light(self.pressed_color)
                    else:
                        self._send_button_color()

        def _send_button_color(self):
            raise NotImplementedError

        def _on_value(self, value, *a, **k):
            if self._notifications_enabled():
                if not self.is_momentary:
                    self._press_button()
                    self._release_button()
                else:
                    if value:
                        self._press_button()
                    else:
                        self._release_button()
                (super(ButtonControlBase.State, self)._on_value)(value, *a, **k)
            self._send_current_color()

        def _press_button(self):
            is_pressed = self._is_pressed
            self._is_pressed = True
            if self._notifications_enabled():
                if not is_pressed:
                    self._on_pressed()

        def _on_pressed(self):
            if self._repeat:
                self._repeat_task.restart()
            self._call_listener("pressed")
            if self._has_delayed_event():
                self._delay_task.restart()
            self._check_double_click_press()

        def _release_button(self):
            is_pressed = self._is_pressed
            self._is_pressed = False
            if self._notifications_enabled():
                if is_pressed:
                    self._on_released()

        def _on_released(self):
            self._call_listener("released")
            if self._repeat:
                self._repeat_task.kill()
            if self._has_delayed_event():
                if self._delay_task.is_running:
                    self._call_listener("released_immediately")
                    self._delay_task.kill()
                else:
                    self._call_listener("released_delayed")
            self._check_double_click_release()

        def _check_double_click_press(self):
            if self._has_listener("double_clicked"):
                if not self._double_click_task.is_running:
                    self._double_click_task.restart()
                    self._double_click_context.click_count = 0
                if self._double_click_context.control_state != self:
                    self._double_click_context.set_new_context(self)

        def _check_double_click_release(self):
            if self._has_listener("double_clicked"):
                if self._double_click_task.is_running:
                    if self._double_click_context.control_state == self:
                        self._double_click_context.click_count += 1
                        if self._double_click_context.click_count == 2:
                            self._call_listener("double_clicked")
                            self._double_click_task.kill()

        def set_double_click_context(self, context):
            self._double_click_context = context

        @lazy_attribute
        def _delay_task(self):
            return self.tasks.add(task.sequence(task.wait(self._delay_time), task.run(self._on_pressed_delayed)))

        @lazy_attribute
        def _repeat_task(self):
            notify_pressed = partial(self._call_listener, "pressed")
            return self.tasks.add(
                task.sequence(
                    task.wait(self._delay_time),
                    task.loop(task.wait(ButtonControlBase.REPEAT_RATE), task.run(notify_pressed)),
                ),
            )

        def _kill_all_tasks(self):
            if self._repeat:
                self._repeat_task.kill()
            if self._has_delayed_event():
                self._delay_task.kill()

        @lazy_attribute
        def _double_click_task(self):
            return self.tasks.add(task.wait(ButtonControlBase.DOUBLE_CLICK_TIME))

        def _has_delayed_event(self):
            return (
                self._has_listener("pressed_delayed")
                or self._has_listener("released_delayed")
                or self._has_listener("released_immediately")
            )

        def _on_pressed_delayed(self):
            if self._is_pressed:
                self._call_listener("pressed_delayed")

        def update(self):
            self._send_current_color()

    def __init__(self, *a, **k):
        super().__init__(extra_args=a, extra_kws=k)


class ButtonControl(ButtonControlBase):
    class State(ButtonControlBase.State):
        color = control_color("DefaultButton.On")

        def __init__(self, color="DefaultButton.On", *a, **k):
            (super(ButtonControl.State, self).__init__)(*a, **k)
            self.color = color

        def _send_button_color(self):
            if self.color is not None:
                self._control_element.set_light(self.color)


class PlayableControl(ButtonControl):
    class Mode(int):
        pass

    Mode.playable = Mode(0)
    Mode.listenable = Mode(1)
    Mode.playable_and_listenable = Mode(2)

    class State(ButtonControl.State):
        def __init__(self, mode=None, *a, **k):
            (super(PlayableControl.State, self).__init__)(*a, **k)
            self._enabled = True
            self._mode = PlayableControl.Mode.playable if mode is None else mode
            self._mode_to_forwarding = {
                PlayableControl.Mode.playable: ScriptForwarding.none,
                PlayableControl.Mode.listenable: ScriptForwarding.exclusive,
                PlayableControl.Mode.playable_and_listenable: ScriptForwarding.non_consuming,
            }

        def set_control_element(self, control_element):
            super(PlayableControl.State, self).set_control_element(control_element)
            self._update_script_forwarding()

        def _update_script_forwarding(self):
            if self._control_element:
                if self._enabled:
                    self._control_element.script_forwarding = self._mode_to_forwarding[self._mode]

        @property
        def enabled(self):
            return self._enabled

        @enabled.setter
        def enabled(self, enabled):
            super(PlayableControl.State, PlayableControl.State).enabled.fset(self, enabled)
            if not enabled or self._control_element:
                self._control_element.reset_state()
                self._send_current_color()
            else:
                self.set_control_element(self._control_element)

        def set_mode(self, value):
            self._mode = value
            self._update_script_forwarding()

        def _is_listenable(self):
            return self._mode != PlayableControl.Mode.playable

        def _notifications_enabled(self):
            return super(PlayableControl.State, self)._notifications_enabled() and self._is_listenable()
